import struct
from copy import deepcopy
from typing import Dict, List, NamedTuple

from pykotor.common.geometry import Vector3, Vector4
from pykotor.common.misc import Game
from pykotor.common.stream import BinaryReader

_GEOM_ROOT_FP0_K1 = 4273776
_GEOM_ROOT_FP1_K1 = 4216096
_GEOM_ROOT_FP0_K2 = 4285200
_GEOM_ROOT_FP1_K2 = 4216320

_GEOM_ANIM_FP0_K1 = 4273392
_GEOM_ANIM_FP1_K1 = 4451552
_GEOM_ANIM_FP0_K2 = 4284816
_GEOM_ANIM_FP1_K2 = 4522928

_NODE_TYPE_MESH = 32
_MESH_HEADER_SIZE_K1 = 332
_MESH_HEADER_SIZE_K2 = 340
_MESH_FP0_K1 = 4216656
_MESH_FP1_K1 = 4216672
_MESH_FP0_K2 = 4216880
_MESH_FP1_K2 = 4216896

_NODE_TYPE_SKIN = 64
_SKIN_HEADER_SIZE = 108
_SKIN_FP0_K1 = 4216592
_SKIN_FP1_K1 = 4216608
_SKIN_FP0_K2 = 4216816
_SKIN_FP1_K2 = 4216832

_NODE_TYPE_DANGLY = 256
_DANGLY_HEADER_SIZE = 28
_DANGLY_FP0_K1 = 4216640
_DANGLY_FP1_K1 = 4216624
_DANGLY_FP0_K2 = 4216864
_DANGLY_FP1_K2 = 4216848

_NODE_TYPE_SABER = 2048
_SABER_HEADER_SIZE = 20
_SABER_FP0_K1 = 4216656
_SABER_FP1_K1 = 4216672
_SABER_FP0_K2 = 4216880
_SABER_FP1_K2 = 4216896

_NODE_TYPE_AABB = 512
_AABB_HEADER_SIZE = 4
_AABB_FP0_K1 = 4216656
_AABB_FP1_K1 = 4216672
_AABB_FP0_K2 = 4216880
_AABB_FP1_K2 = 4216896

_NODE_TYPE_LIGHT = 2
_LIGHT_HEADER_SIZE = 92

_NODE_TYPE_EMITTER = 4
_EMITTER_HEADER_SIZE = 224


class MDLMDXTuple(NamedTuple):
    mdl: bytes
    mdx: bytes


def rename(
        data: bytes,
        name: str
) -> bytes:
    return data[:20] + name.ljust(32, '\0').encode('ascii') + data[52:]


def list_textures(
        data: bytes
) -> List[str]:
    textures = []

    with BinaryReader.from_bytes(data, 12) as reader:
        reader.seek(168)
        root_offset = reader.read_uint32()

        offsets = {}
        nodes = [root_offset]
        while nodes:
            node_offset = nodes.pop()
            reader.seek(node_offset)
            node_id = reader.read_uint32()

            reader.seek(node_offset + 44)
            child_offsets_offset = reader.read_uint32()
            child_offsets_count = reader.read_uint32()

            reader.seek(child_offsets_offset)
            for i in range(child_offsets_count):
                nodes.append(reader.read_uint32())

            if node_id & 32:
                reader.seek(node_offset + 168)
                texture = reader.read_string(32)
                if texture != "" and texture != "NULL" and texture.lower() not in textures:
                    textures.append(texture.lower())

    return textures


def list_lightmaps(
        data: bytes
) -> List[str]:
    lightmaps = []

    with BinaryReader.from_bytes(data, 12) as reader:
        reader.seek(168)
        root_offset = reader.read_uint32()

        offsets = {}
        nodes = [root_offset]
        while nodes:
            node_offset = nodes.pop()
            reader.seek(node_offset)
            node_id = reader.read_uint32()

            reader.seek(node_offset + 44)
            child_offsets_offset = reader.read_uint32()
            child_offsets_count = reader.read_uint32()

            reader.seek(child_offsets_offset)
            for i in range(child_offsets_count):
                nodes.append(reader.read_uint32())

            if node_id & 32:
                reader.seek(node_offset + 200)
                lightmap = reader.read_string(32)
                if lightmap != "" and lightmap != "NULL" and lightmap.lower() not in lightmaps:
                    lightmaps.append(lightmap.lower())

    return lightmaps


def change_textures(
        data: bytes,
        textures: Dict[str, str]
) -> bytes:
    data = bytearray(data)
    offsets = {}

    textures_ins = {}
    for old_texture, new_texture in textures.items():
        textures_ins[old_texture.lower()] = new_texture.lower()
    textures = textures_ins

    with BinaryReader.from_bytes(data, 12) as reader:
        reader.seek(168)
        root_offset = reader.read_uint32()

        nodes = [root_offset]
        while nodes:
            node_offset = nodes.pop()
            reader.seek(node_offset)
            node_id = reader.read_uint32()

            reader.seek(node_offset + 44)
            child_offsets_offset = reader.read_uint32()
            child_offsets_count = reader.read_uint32()

            reader.seek(child_offsets_offset)
            for i in range(child_offsets_count):
                nodes.append(reader.read_uint32())

            if node_id & 32:
                reader.seek(node_offset + 168)
                texture = reader.read_string(32).lower()

                if texture in textures:
                    if texture in offsets:
                        offsets[texture].append(node_offset + 168)
                    else:
                        offsets[texture] = [node_offset + 168]

        for texture, offsets in offsets.items():
            for offset in offsets:
                offset += 12
                data = data[:offset] + struct.pack('32s', textures[texture].ljust(32, '\0').encode('ascii')) + data[
                                                                                                               offset + 32:]

    return bytes(data)


def change_lightmaps(
        data: bytes,
        textures: Dict[str, str]
) -> bytes:
    data = bytearray(data)
    offsets = {}

    textures_ins = {}
    for old_texture, new_texture in textures.items():
        textures_ins[old_texture.lower()] = new_texture.lower()
    textures = textures_ins

    with BinaryReader.from_bytes(data, 12) as reader:
        reader.seek(168)
        root_offset = reader.read_uint32()

        nodes = [root_offset]
        while nodes:
            node_offset = nodes.pop()
            reader.seek(node_offset)
            node_id = reader.read_uint32()

            reader.seek(node_offset + 44)
            child_offsets_offset = reader.read_uint32()
            child_offsets_count = reader.read_uint32()

            reader.seek(child_offsets_offset)
            for i in range(child_offsets_count):
                nodes.append(reader.read_uint32())

            if node_id & 32:
                reader.seek(node_offset + 200)
                texture = reader.read_string(32).lower()

                if texture in textures:
                    if texture in offsets:
                        offsets[texture].append(node_offset + 200)
                    else:
                        offsets[texture] = [node_offset + 200]

        for texture, offsets in offsets.items():
            for offset in offsets:
                offset += 12
                data = data[:offset] + struct.pack('32s', textures[texture].ljust(32, '\0').encode('ascii')) + data[
                                                                                                               offset + 32:]

    return bytes(data)


def detect_version(
        data: bytes
) -> Game:
    pointer = struct.unpack("I", data[12:16])[0]
    return Game.K1 if pointer == _GEOM_ROOT_FP0_K1 else Game.K2


def convert_to_k1(
        data: bytes
) -> bytes:
    if detect_version(data) == Game.K1:
        return data

    trim = []

    with BinaryReader.from_bytes(data, 12) as reader:
        reader.seek(168)
        root_offset = reader.read_uint32()

        nodes = [root_offset]
        while nodes:
            node_offset = nodes.pop()
            reader.seek(node_offset)
            node_type = reader.read_uint16()

            if node_type & 32:
                trim.append((node_type, node_offset))

            reader.seek(node_offset + 44)
            child_offsets_offset = reader.read_uint32()
            child_offsets_count = reader.read_uint32()

            reader.seek(child_offsets_offset)
            for i in range(child_offsets_count):
                nodes.append(reader.read_uint32())

    start = data[:12]
    data = bytearray(data[12:])

    data[0:4] = struct.pack("I", _GEOM_ROOT_FP0_K1)
    data[4:8] = struct.pack("I", _GEOM_ROOT_FP1_K1)

    # TODO Animations

    for node_type, node_offset in trim:
        mesh_start = node_offset + 80  # Start of mesh header

        offset_start = node_offset + 80 + 332  # Location of start of bytes to be shifted
        offset_size = 8  # How many bytes we have to shift

        if node_type & _NODE_TYPE_SKIN:
            offset_size += _SKIN_HEADER_SIZE
            data[mesh_start:mesh_start + 4] = struct.pack("I", _MESH_FP0_K1)
            data[mesh_start + 4:mesh_start + 8] = struct.pack("I", _MESH_FP1_K1)

        if node_type & _NODE_TYPE_DANGLY:
            offset_size += _DANGLY_HEADER_SIZE
            data[mesh_start:mesh_start + 4] = struct.pack("I", _DANGLY_FP0_K1)
            data[mesh_start + 4:mesh_start + 8] = struct.pack("I", _DANGLY_FP1_K1)

        if node_type & _NODE_TYPE_SABER:
            offset_size += _SABER_HEADER_SIZE
            data[mesh_start:mesh_start + 4] = struct.pack("I", _SABER_FP0_K1)
            data[mesh_start + 4:mesh_start + 8] = struct.pack("I", _SABER_FP1_K1)

        if node_type & _NODE_TYPE_AABB:
            offset_size += _AABB_HEADER_SIZE
            data[mesh_start:mesh_start + 4] = struct.pack("I", _AABB_FP0_K1)
            data[mesh_start + 4:mesh_start + 8] = struct.pack("I", _AABB_FP1_K1)

        shifting = data[offset_start:offset_start + offset_size]
        data[offset_start - 8:offset_start - 8 + offset_size] = shifting

    return bytes(start + data)


def convert_to_k2(
        data: bytes
) -> bytes:
    if detect_version(data) == Game.K2:
        return data

    offsets = {}  # Maps the offset for an offset to its offset
    mesh_offsets = []  # Tuple of (Offset to every mesh node, Node type)
    anim_offsets = []

    # First, we build a dictionary of every offset in the file plus a list of the mesh nodes
    with BinaryReader.from_bytes(data, 12) as reader:
        def node_recursive(
                offset_to_root_offset
        ):
            nodes = [offset_to_root_offset]
            while nodes:
                offset_to_node_offset = nodes.pop()
                reader.seek(offset_to_node_offset)
                node_offset = reader.read_uint32()
                offsets[
                    offset_to_node_offset] = node_offset  # Geometry header/Node children offsets array -> Node offset

                reader.seek(node_offset)
                node_type = reader.read_uint16()

                base_offset = node_offset + 80
                if node_type & _NODE_TYPE_MESH:
                    mesh_offsets.append([node_offset, node_type])

                    reader.seek(base_offset + 8)
                    offsets[base_offset + 8] = reader.read_uint32()  # Node header -> Face array offset

                    reader.seek(base_offset + 176)
                    offsets[base_offset + 176] = reader.read_uint32()  # Node header -> Vertex indices count array
                    indicies_array_count = reader.read_uint32()

                    reader.seek(base_offset + 188)
                    offsets[base_offset + 188] = reader.read_uint32()  # Node header -> Vertex indices locations array
                    if indicies_array_count == 1:
                        indices_locations_offset = offsets[base_offset + 188]
                        reader.seek(indices_locations_offset)
                        offsets[
                            indices_locations_offset] = reader.read_uint32()  # Vertex indices locations array -> Vertex indices array

                    reader.seek(base_offset + 200)
                    offsets[base_offset + 200] = reader.read_uint32()  # Node header -> Inverted counter array

                    reader.seek(base_offset + 328)
                    offsets[base_offset + 328] = reader.read_uint32()  # Node header -> Vertex array

                    base_offset += _MESH_HEADER_SIZE_K1

                if node_type & _NODE_TYPE_LIGHT:
                    reader.seek(base_offset + 4)
                    offsets[base_offset + 4] = reader.read_uint32()  # Node header -> Lens flare size array

                    reader.seek(base_offset + 16)
                    offsets[base_offset + 16] = reader.read_uint32()  # Node header -> Lens flare size array

                    reader.seek(base_offset + 28)
                    offsets[base_offset + 28] = reader.read_uint32()  # Node header -> Lens flare positions array

                    reader.seek(base_offset + 40)
                    offsets[base_offset + 40] = reader.read_uint32()  # Node header -> Flare colour shifts array

                    reader.seek(base_offset + 52)
                    offsets[base_offset + 52] = reader.read_uint32()  # Node header -> Flare texture names offsets array
                    texture_count = reader.read_uint32()

                    for i in range(texture_count):
                        reader.seek(offsets[base_offset + 52] + i * 4)
                        offsets[offsets[base_offset + 52] + i * 4] = reader.read_uint32()

                if node_type & _NODE_TYPE_EMITTER:
                    ...  # No offsets. Hooray.

                if node_type & _NODE_TYPE_SKIN:
                    reader.seek(base_offset + 20)
                    offsets[base_offset + 20] = reader.read_uint32()  # Node header -> Bone map array

                    reader.seek(base_offset + 28)
                    offsets[base_offset + 28] = reader.read_uint32()  # Node header -> QBones array

                    reader.seek(base_offset + 40)
                    offsets[base_offset + 40] = reader.read_uint32()  # Node header -> TBones Array

                    reader.seek(base_offset + 52)
                    offsets[base_offset + 52] = reader.read_uint32()  # Node header -> Array8

                    base_offset += _SKIN_HEADER_SIZE

                if node_type & _NODE_TYPE_DANGLY:
                    reader.seek(base_offset + 0)
                    offsets[base_offset + 0] = reader.read_uint32()  # Node header -> Dangly constraint array

                    reader.seek(base_offset + 24)
                    offsets[base_offset + 24] = reader.read_uint32()  # Node header -> Unknown

                    base_offset += _DANGLY_HEADER_SIZE

                if node_type & _NODE_TYPE_AABB:
                    reader.seek(base_offset + 0)
                    offsets[base_offset + 0] = reader.read_uint32()  # Node header -> AABB root node

                    aabbs = [offsets[base_offset + 0]]
                    while aabbs:
                        aabb = aabbs.pop()

                        reader.seek(aabb + 24)
                        leaf0 = reader.read_uint32()
                        if leaf0:
                            aabbs.append(leaf0)
                            offsets[aabb + 24] = leaf0  # AABB Node -> AABB child node

                        reader.seek(aabb + 28)
                        leaf1 = reader.read_uint32()
                        if leaf1:
                            aabbs.append(leaf1)
                            offsets[aabb + 28] = leaf1  # AABB Node -> AABB child node

                    base_offset += _AABB_HEADER_SIZE

                if node_type & _NODE_TYPE_SABER:
                    reader.seek(base_offset + 0)
                    offsets[base_offset + 0] = reader.read_uint32()  # Node header -> Saber Verts array

                    reader.seek(base_offset + 4)
                    offsets[base_offset + 4] = reader.read_uint32()  # Node header -> Saber UVs array

                    reader.seek(base_offset + 8)
                    offsets[base_offset + 8] = reader.read_uint32()  # Node header -> Saber Normals array

                    base_offset += _SABER_HEADER_SIZE

                reader.seek(node_offset + 8)
                offsets[node_offset + 8] = reader.read_uint32()  # Node header -> Geometry header

                reader.seek(node_offset + 12)
                offsets[node_offset + 12] = reader.read_uint32()  # Node header -> Parent node

                reader.seek(node_offset + 56)
                offsets[node_offset + 56] = reader.read_uint32()  # Node header -> controller array offset

                reader.seek(node_offset + 68)
                offsets[node_offset + 68] = reader.read_uint32()  # Node header -> controller data offset

                reader.seek(node_offset + 44)
                child_offsets_offset = reader.read_uint32()
                child_offsets_count = reader.read_uint32()
                offsets[node_offset + 44] = child_offsets_offset  # Node header -> Child node offsets array

                for i in range(child_offsets_count):
                    nodes.append(child_offsets_offset + i * 4)

        reader.seek(88)
        anim_locations_offset = reader.read_uint32()
        anim_count = reader.read_uint32()

        reader.seek(168)
        root_offset = reader.read_uint32()
        node_count = reader.read_uint32()

        reader.seek(184)
        name_offsets = reader.read_uint32()
        name_count = reader.read_uint32()

        reader.seek(40)
        offsets[40] = reader.read_uint32()  # Model header -> Root node

        reader.seek(88)
        offsets[88] = reader.read_uint32()  # Model header -> Animation array

        reader.seek(168)
        offsets[168] = reader.read_uint32()  # Model header -> Head root

        reader.seek(184)
        offsets[184] = reader.read_uint32()  # Model header -> Name offsets array

        for i in range(name_count):
            reader.seek(offsets[184] + i * 4)
            offsets[offsets[184] + i * 4] = reader.read_uint32()  # Name offset array -> Name

        for i in range(anim_count):
            reader.seek(anim_locations_offset + i * 4)
            animation_offset = reader.read_uint32()
            offsets[anim_locations_offset + i * 4] = animation_offset  # Event locations array -> Offset to event
            anim_offsets.append(animation_offset)

            reader.seek(animation_offset + 120)
            offsets[animation_offset + 120] = reader.read_uint32()  # Animation header -> Offset to event array

            reader.seek(animation_offset + 40)
            node_recursive(animation_offset + 40)

        node_recursive(168)

    # Second, we will update the function pointers to use K2 values instead of K1
    mdx_size = data[8:12]
    data = bytearray(data[12:])

    data[0:4] = struct.pack("I", _GEOM_ROOT_FP0_K2)
    data[4:8] = struct.pack("I", _GEOM_ROOT_FP1_K2)

    for anim_offset in anim_offsets:
        data[anim_offset:anim_offset + 4] = struct.pack("I", _GEOM_ANIM_FP0_K2)
        data[anim_offset + 4:anim_offset + 8] = struct.pack("I", _GEOM_ANIM_FP1_K2)

    for node_offset, node_type in mesh_offsets:
        mesh_start = node_offset + 80  # Start of mesh header

        if node_type & _NODE_TYPE_SKIN:
            data[mesh_start:mesh_start + 4] = struct.pack("I", _MESH_FP0_K2)
            data[mesh_start + 4:mesh_start + 8] = struct.pack("I", _MESH_FP1_K2)

        if node_type & _NODE_TYPE_DANGLY:
            data[mesh_start:mesh_start + 4] = struct.pack("I", _DANGLY_FP0_K2)
            data[mesh_start + 4:mesh_start + 8] = struct.pack("I", _DANGLY_FP1_K2)

        if node_type & _NODE_TYPE_SABER:
            data[mesh_start:mesh_start + 4] = struct.pack("I", _SABER_FP0_K1)
            data[mesh_start + 4:mesh_start + 8] = struct.pack("I", _SABER_FP1_K2)

        if node_type & _NODE_TYPE_AABB:
            data[mesh_start:mesh_start + 4] = struct.pack("I", _AABB_FP0_K2)
            data[mesh_start + 4:mesh_start + 8] = struct.pack("I", _AABB_FP1_K2)

    offsets = {k: v for k, v in sorted(list(offsets.items()), reverse=True)}

    # Third, we will add the extra data in the mesh headers and update all offsets in our dictionary
    shifted = 0
    for i in range(len(mesh_offsets)):
        node_offset, node_type = mesh_offsets[i]
        insert_location = node_offset + 80 + 324
        data = data[:insert_location] + bytes([0 for i in range(8)]) + data[insert_location:]

        for offset_location, offset_value in deepcopy(offsets).items():
            if insert_location < offset_location:
                del offsets[offset_location]
                if offset_location + 8 in offsets:
                    raise ValueError("whoops")
                offsets[offset_location + 8] = offset_value

        for offset_location, offset_value in deepcopy(offsets).items():
            if insert_location < offset_value:
                offsets[offset_location] = offset_value + 8

        for j in range(len(mesh_offsets)):
            if insert_location < mesh_offsets[j][0]:
                mesh_offsets[j][0] += 8

        shifted += 8

    # Finally, we update the offsets in the bytes with the offsets in our dictionary
    for offset_location, offset_value in offsets.items():
        data[offset_location:offset_location + 4] = struct.pack("I", offset_value)

    return bytes([0 for i in range(4)]) + struct.pack("I", len(data)) + mdx_size + data


def transform(
        data: bytes,
        translation: Vector3,
        rotation: float
) -> bytes:
    """
    Performs a translation and then rotation on the target MDL data.

    Args:
        data: MDL data.
        translation: Translation value.
        rotation: Rotation value.

    Returns:
        The MDL data post-transformation.
    """
    orientation = Vector4.from_euler(0, 0, rotation)
    mdx_size = struct.unpack("I", data[8:12])[0]
    data = bytearray(data[12:])

    with BinaryReader.from_bytes(data) as reader:
        reader.seek(44)
        node_count = reader.read_uint32()

        reader.seek(168)
        root_offset = reader.read_uint32()

        reader.seek(root_offset)
        type_id = reader.read_uint16()
        supernode_id = reader.read_uint16()
        label_id = reader.read_uint32()
        reader.skip(6)
        reader.skip(4)
        reader.skip(4)
        reader.skip(4 * 3)
        reader.skip(4 * 4)

        reader.seek(root_offset + 44)
        child_array_offset = reader.read_uint32()
        child_count = reader.read_uint32()

    if child_count == 0:
        return data

    root_child_array_offset = len(data)
    insert_node_offset = len(data) + 4
    insert_controller_offset = insert_node_offset + 80
    insert_controller_data_offset = insert_controller_offset + 32

    data[44:48] = struct.pack("I", node_count + 1)

    data[root_offset+44:root_offset+48] = struct.pack("I", root_child_array_offset)
    data[root_offset+48:root_offset+52] = struct.pack("I", 1)
    data[root_offset+52:root_offset+56] = struct.pack("I", 1)

    data += struct.pack("I", insert_node_offset)
    data += struct.pack("HHHH II fff ffff III III III",
                        1,                      # Node Type
                        node_count + 1,         # Node ID
                        1,                      # Label ID (steal some existing node's label)
                        0,                      # Padding

                        0,
                        root_offset,
                        *translation,              # Node Position
                        orientation.w, orientation.x, orientation.y, orientation.z,           # Node Orientation
                        child_array_offset,     # Child Array Offset
                        child_count,
                        child_count,
                        insert_controller_offset,  # Controller Array
                        2,
                        2,
                        insert_controller_data_offset,  # Controller Data Array
                        9,
                        9
    )

    data += struct.pack("IHHHHBBBB", 8, 0xFFFF, 1, 0, 1, 3, 0, 0, 0)
    data += struct.pack("IHHHHBBBB", 20, 0xFFFF, 1, 4, 5, 4, 0, 0, 0)
    data += struct.pack("ffff", 0.0, *translation)
    data += struct.pack("fffff", 0.0, *orientation)

    return struct.pack("III", 0, len(data), mdx_size) + data


def flip(
        mdl_data: bytes,
        mdx_data: bytes,
        flip_x: bool,
        flip_y: bool
) -> MDLMDXTuple:
    """
    Returns the given MDL and MDX data with the vertices flipped along the specified axes.

    Args:
        mdl_data: The raw MDL data.
        mdx_data: The raw MDX data.
        flip_x: Flip the vertices across the X-axis.
        flip_y: Flip the vertices across the Y-axis.

    Returns:
        The MDL and MDX data post-flip.
    """

    # If neither bools are set to True, no transformations need to be done and we can just return the original data
    if not flip_x and not flip_y:
        return MDLMDXTuple(mdl_data, mdx_data)

    # The data we need to change:
    #    1. The vertices stored in the MDL
    #    2. The vertex positions, normals, stored in the MDX

    # Trim the data to correct the offsets
    mdl_start = mdl_data[:12]
    mdl_data = bytearray(mdl_data[12:])
    mdx_data = bytearray(mdx_data)

    mdl_vertex_offsets = []  # This is a list of tuples: (count, offset)
    mdx_vertex_offsets = []  # This is a list of tuples: (count, offset, stride, position)
    mdx_normal_offsets = []  # This is a list of tuples: (count, offset, stride, position)
    elements_offsets = []  # This is a list of tuples: (count, offset)
    faces_offsets = []  # This is a list of tuples: (count, offset)
    with BinaryReader.from_bytes(mdl_data) as reader:
        reader.seek(168)
        root_offset = reader.read_uint32()

        nodes = [root_offset]
        while nodes:
            node_offset = nodes.pop()
            reader.seek(node_offset)
            node_id = reader.read_uint32()

            mdl_vertex_offsets.append((1, node_offset+16))

            # Need to determine the location of the position controller
            reader.seek(node_offset + 56)
            controllers_offset = reader.read_uint32()
            controllers_count = reader.read_uint32()

            reader.seek(node_offset + 68)
            controller_datas_offset = reader.read_uint32()
            controller_datas_count = reader.read_uint32()

            for i in range(controllers_count):
                reader.seek(controllers_offset + i*16)
                controller_type = reader.read_uint32()
                if controller_type == 8:
                    reader.skip(6)
                    data_offset = reader.read_uint16()
                    mdl_vertex_offsets.append((1, controller_datas_offset + data_offset*4))

            reader.seek(node_offset + 44)
            child_offsets_offset = reader.read_uint32()
            child_offsets_count = reader.read_uint32()

            reader.seek(child_offsets_offset)
            for i in range(child_offsets_count):
                nodes.append(reader.read_uint32())

            if node_id & 32:
                reader.seek(node_offset+80)
                fp = reader.read_uint32()
                tsl = False if fp in [_MESH_FP0_K1, _SKIN_FP0_K1, _DANGLY_FP0_K2, _AABB_FP0_K1, _SABER_FP0_K1] else True

                reader.seek(node_offset+80+8)
                faces_offset = reader.read_uint32()
                faces_count = reader.read_uint32()
                faces_offsets.append((faces_count, faces_offset))

                reader.seek(node_offset+80+188)
                offset_to_elements_offset = reader.read_uint32()
                reader.seek(offset_to_elements_offset)
                elements_offset = reader.read_uint32()
                elements_offsets.append((faces_count, elements_offset))

                reader.seek(node_offset+80+304)
                vertex_count = reader.read_uint16()
                reader.seek(node_offset+80+336 if tsl else node_offset+80+328)
                vertex_offset = reader.read_uint32()
                mdl_vertex_offsets.append((vertex_count, vertex_offset))

                reader.seek(node_offset+80+252)
                mdx_stride = reader.read_uint32()
                mdx_bitmap = reader.read_uint32()
                reader.seek(node_offset+80+260)
                mdx_offset_pos = reader.read_uint32()
                mdx_offset_norm = reader.read_uint32()
                reader.seek(node_offset+80+332 if tsl else node_offset+80+324)
                mdx_start = reader.read_uint32()
                mdx_vertex_offsets.append((vertex_count, mdx_start, mdx_stride, mdx_offset_pos))
                mdx_normal_offsets.append((vertex_count, mdx_start, mdx_stride, mdx_offset_norm))

    # Fix vertex order
    if flip_x != flip_y:
        for count, start_offset in elements_offsets:
            for i in range(count):
                offset = start_offset + i*6
                v1 = struct.unpack("H", mdl_data[offset:offset+2])[0]
                v2 = struct.unpack("H", mdl_data[offset+2:offset+4])[0]
                v3 = struct.unpack("H", mdl_data[offset+4:offset+6])[0]
                mdl_data[offset:offset+2] = struct.pack("H", v1)
                mdl_data[offset+2:offset+4] = struct.pack("H", v3)
                mdl_data[offset+4:offset+6] = struct.pack("H", v2)
        for count, start_offset in faces_offsets:
            for i in range(count):
                offset = start_offset + i*32 + 26
                v1 = struct.unpack("H", mdl_data[offset:offset+2])[0]
                v2 = struct.unpack("H", mdl_data[offset+2:offset+4])[0]
                v3 = struct.unpack("H", mdl_data[offset+4:offset+6])[0]
                mdl_data[offset:offset+2] = struct.pack("H", v1)
                mdl_data[offset+2:offset+4] = struct.pack("H", v3)
                mdl_data[offset+4:offset+6] = struct.pack("H", v2)

    # Update the MDL vertices
    for count, start_offset in mdl_vertex_offsets:
        for i in range(count):
            offset = start_offset + i*12
            if flip_x:
                x = struct.unpack("f", mdl_data[offset:offset+4])[0]
                mdl_data[offset:offset+4] = struct.pack("f", -x)
            if flip_y:
                y = struct.unpack("f", mdl_data[offset+4:offset+8])[0]
                mdl_data[offset+4:offset+8] = struct.pack("f", -y)

    # Update the MDX vertices
    for count, start_offset, stride, position in mdx_vertex_offsets:
        for i in range(count):
            offset = start_offset + i*stride + position
            if flip_x:
                x = struct.unpack("f", mdx_data[offset:offset+4])[0]
                mdx_data[offset:offset+4] = struct.pack("f", -x)
            if flip_y:
                y = struct.unpack("f", mdx_data[offset+4:offset+8])[0]
                mdx_data[offset+4:offset+8] = struct.pack("f", -y)

    # Update the MDX normals
    for count, start_offset, stride, position in mdx_normal_offsets:
        for i in range(count):
            offset = start_offset + i * stride + position
            if flip_x:
                x = struct.unpack("f", mdx_data[offset:offset + 4])[0]
                mdx_data[offset:offset + 4] = struct.pack("f", -x)
            if flip_y:
                y = struct.unpack("f", mdx_data[offset + 4:offset + 8])[0]
                mdx_data[offset + 4:offset + 8] = struct.pack("f", -y)

    # Readd the first 12 bytes
    mdl_data = mdl_start + mdl_data

    return MDLMDXTuple(mdl_data, mdx_data)



from unittest import TestCase

from pykotor.common.misc import Game
from pykotor.resource.formats.gff import read_gff
from pykotor.resource.generics.dlg import construct_dlg, dismantle_dlg

TEST_FILE = "../../files/test.dlg"


class TestDLG(TestCase):
    def test_io(self):
        gff = read_gff(TEST_FILE)
        dlg = construct_dlg(gff)
        self.validate_io(dlg)

        gff = dismantle_dlg(dlg, Game.K2)
        dlg = construct_dlg(gff)
        self.validate_io(dlg)

    def validate_io(self, dlg):
        all_entries = dlg.all_entries()
        all_replies = dlg.all_replies()

        entry0 = all_entries[0]
        entry1 = all_entries[1]
        entry2 = all_entries[2]

        reply0 = all_replies[0]
        reply1 = all_replies[1]

        self.assertEqual(3, len(all_entries))
        self.assertEqual(2, len(all_replies))
        self.assertEqual(2, len(dlg.starters))
        self.assertEqual(2, len(dlg.stunts))

        self.assertIn(entry0, [link.node for link in dlg.starters])
        self.assertIn(entry2, [link.node for link in dlg.starters])

        self.assertEqual(2, len(entry0.links))
        self.assertIn(reply0, [link.node for link in entry0.links])
        self.assertIn(reply1, [link.node for link in entry0.links])

        self.assertEqual(1, len(reply0.links))
        self.assertIn(entry0, [link.node for link in reply0.links])

        self.assertEqual(1, len(reply1.links))
        self.assertIn(entry1, [link.node for link in reply1.links])

        self.assertEqual(0, len(entry2.links))

        self.assertEqual(13, dlg.delay_entry)
        self.assertEqual(14, dlg.delay_reply)
        self.assertEqual(1337, dlg.word_count)
        self.assertEqual("abort", dlg.on_abort)
        self.assertEqual("end", dlg.on_end)
        self.assertEqual(1, dlg.skippable)
        self.assertEqual("track", dlg.ambient_track)
        self.assertEqual(123, dlg.animated_cut)
        self.assertEqual("camm", dlg.camera_model)
        self.assertEqual(1, dlg.computer_type.value)
        self.assertEqual(1, dlg.conversation_type.value)
        self.assertEqual(1, dlg.old_hit_check)
        self.assertEqual(1, dlg.unequip_hands)
        self.assertEqual(1, dlg.unequip_items)
        self.assertEqual("echo", dlg.vo_id)
        self.assertEqual(123, dlg.alien_race_owner)
        self.assertEqual(12, dlg.post_proc_owner)
        self.assertEqual(3, dlg.record_no_vo)

        self.assertEqual("yoohoo", entry0.listener)
        self.assertEqual(-1, entry0.text.stringref)
        self.assertEqual("gand", entry0.vo_resref)
        self.assertEqual("num1", entry0.script1)
        self.assertEqual(-1, entry0.delay)
        self.assertEqual("commentto", entry0.comment)
        self.assertEqual("gonk", entry0.sound)
        self.assertEqual("quest", entry0.quest)
        self.assertEqual(-1, entry0.plot_index)
        self.assertEqual(1.0, entry0.plot_xp_percentage)
        self.assertEqual(1, entry0.wait_flags)
        self.assertEqual(14, entry0.camera_angle)
        self.assertEqual(1, entry0.fade_type)
        self.assertEqual(1, entry0.sound_exists)
        self.assertEqual(1, entry0.alien_race_node)
        self.assertEqual(1, entry0.vo_text_changed)
        self.assertEqual(4, entry0.emotion_id)
        self.assertEqual(2, entry0.facial_id)
        self.assertEqual(1, entry0.node_id)
        self.assertEqual(1, entry0.unskippable)
        self.assertEqual(3, entry0.post_proc_node)
        self.assertEqual(1, entry0.record_vo)
        self.assertEqual("num2", entry0.script2)
        self.assertEqual(1, entry0.vo_text_changed)
        self.assertEqual(1, entry0.record_no_vo_override)
        self.assertEqual(32, entry0.camera_id)
        self.assertEqual("bark", entry0.speaker)
        self.assertEqual(-1, entry0.camera_effect)
        self.assertEqual(1, entry0.record_no_vo_override)

        self.assertEqual("bbb", dlg.stunts[1].participant)
        self.assertEqual("m01aa_c04_char01", dlg.stunts[1].stunt_model)



from __future__ import annotations
from enum import Enum
from typing import List, Optional, Any


class ScriptConstant:
    def __init__(
            self,
            datatype: DataType,
            name: str,
            value: Any
    ):
        self.datatype: DataType = DataType(datatype)
        self.name: str = name
        self.value: Any = value

        if self.datatype == DataType.INT and not isinstance(value, int):
            raise ValueError("Script constant value argument does not match given datatype.")
        elif self.datatype == DataType.FLOAT and not isinstance(value, float):
            raise ValueError("Script constant value argument does not match given datatype.")
        elif self.datatype == DataType.STRING and not isinstance(value, str):
            raise ValueError("Script constant value argument does not match given datatype.")

    def __repr__(
            self
    ):
        return 'ScriptConstant("{}", "{}", "{}")'.format(self.datatype, self.name, self.value)

    def __str__(
            self
    ):
        return "{} {} = {};".format(self.datatype, self.name, self.value)


class ScriptParam:
    def __init__(
            self,
            datatype: DataType,
            name: str,
            default: Optional[Any]
    ):
        self.datatype: DataType = DataType(datatype)
        self.name: str = name
        self.default: Optional[str] = default

    def __repr__(
            self
    ):
        return 'ScriptParam({}, {}, {})'.format(repr(self.datatype), repr(self.name), repr(self.default))

    def __str__(
            self
    ):
        if self.default is not None:
            return "{} {} = {}".format(self.datatype, self.name, self.default)
        else:
            return "{} {}".format(self.datatype, self.name)


class ScriptFunction:
    def __init__(
            self,
            returntype: DataType,
            name: str,
            params: List[ScriptParam],
            description: str,
            raw: str
    ):
        self.returntype: DataType = DataType(returntype)
        self.name: str = name
        self.params: List[ScriptParam] = params
        self.description: str = description
        self.raw: str = raw

    def __repr__(
            self
    ):
        return "ScriptFunction({}, {}, {}, {}, {})".format(repr(self.returntype), repr(self.name), repr(self.params),
                                                           repr(self.description), repr(self.raw))

    def __str__(
            self
    ):
        param_str = ""
        for param in self.params:
            param_str += str(param)
            if param is not self.params[-1]:
                param_str += ", "
        return "{} {}({})".format(self.returntype, self.name, param_str)


class DataType(Enum):
    VOID = "void"
    INT = "int"
    FLOAT = "float"
    STRING = "string"
    OBJECT = "object"
    VECTOR = "vector"
    LOCATION = "location"
    EVENT = "event"
    EFFECT = "effect"
    ITEMPROPERTY = "itemproperty"
    TALENT = "talent"
    ACTION = "action"
    STRUCT = "struct"

    def size(self) -> int:
        if self == DataType.VOID:
            return 0
        elif self == DataType.VECTOR:
            return 12
        elif self == DataType.STRUCT:
            raise ValueError  # TODO
        else:
            return 4



import os

from pykotor.extract.installation import Installation
from pykotor.resource.formats.tlk import read_tlk
from pykotor.resource.generics.dlg import read_dlg
from pykotor.resource.type import ResourceType

k1_path = "C:/Program Files (x86)/Steam/steamapps/common/swkotor/"
k2_path = "C:/Users/hugin/Documents/Apps/TSLVanilla/"
k1_inst = Installation(k1_path)
k2_inst = Installation(k2_path)
k1_tlk = read_tlk(k1_path + "dialog.tlk")
k2_tlk = read_tlk(k2_path + "dialog.tlk")

k1_voice = "C:/Users/hugin/Documents/tts/hk/k1"
k2_voice = "C:/Users/hugin/Documents/tts/hk/k2"

vo = {}

k1_resources = k1_inst.chitin_resources()
for module in k1_inst.modules_list():
    k1_resources.extend(k1_inst.module_resources(module))

k2_resources = k2_inst.chitin_resources()
for module in k2_inst.modules_list():
    k2_resources.extend(k2_inst.module_resources(module))

for file in os.listdir(k1_voice) + os.listdir(k2_voice):
    file = file.lower().replace(".wav", "")
    vo[file] = ""

for res in k1_resources:
    if res.restype() == ResourceType.DLG:
        dlg = read_dlg(res.data())

        for entry in dlg.all_entries():
            file = entry.vo_resref.get().lower()
            if file in vo:
                english = k1_tlk.get(entry.text.stringref)
                if english is not None and not english.text.isspace():
                    vo[file] = english.text
            elif entry.speaker.lower() == "hk47":
                print("add", file)
                vo[file] = None



import os
import shutil

from pykotor.common.stream import BinaryReader
from pykotor.tools import sound

from pykotor.extract.installation import Installation
from pykotor.resource.formats.ssf import read_ssf, SSFSound
from pykotor.resource.formats.tlk import read_tlk
from pykotor.resource.generics.dlg import read_dlg
from pykotor.resource.type import ResourceType

k1_path = "C:/Program Files (x86)/Steam/steamapps/common/swkotor/"
k2_path = "C:/Users/hugin/Documents/Apps/TSLVanilla/"
k1_inst = Installation(k1_path)
k2_inst = Installation(k2_path)
k1_tlk = read_tlk(k1_path + "dialog.tlk")
k2_tlk = read_tlk(k2_path + "dialog.tlk")

voice_folder = "C:/Users/hugin/Documents/tts/hk/merge"
voices = {}

files = {}
for path in [os.path.join(path, name) for path, subdirs, files in os.walk(k1_inst.streamvoice_path()) for name in files] \
        + [os.path.join(path, name) for path, subdirs, files in os.walk(k1_inst.streamsounds_path()) for name in files]:
    file = os.path.basename(path).replace(".wav", "").lower()
    files[file] = path


def search(inst, tlk, dlg_tags, ssf_files, dlg_files):
    resources = inst.chitin_resources()
    for module in inst.modules_list():
        resources.extend(inst.module_resources(module))

    # Soundset files
    for res in resources:
        if res.restype() != ResourceType.SSF:
            continue
        if res.resname().lower() not in ssf_files:
            continue

        ssf = read_ssf(res.data())
        for i in range(28):
            sound = ssf.get(SSFSound(i))
            if sound != -1 and tlk.get(sound).text != "" and tlk.get(sound).voiceover.get() != "":
                voices[tlk.get(sound).voiceover.get().lower()] = tlk.get(sound).text

    # Dialog files
    for res in resources:
        if res.restype() != ResourceType.DLG:
            continue

        dlg = read_dlg(res.data())

        for entry in dlg.all_entries():
            file = entry.vo_resref.get().lower()

            if entry.speaker.lower() in dlg_tags or res.resname().lower() in dlg_files:
                english = tlk.get(entry.text.stringref)
                if english is not None and not english.text.isspace():
                    voices[file] = english.text


search(k1_inst, k1_tlk, ["hk47"], ["p_hk47"], ["k_hhkd_dialog"])
for vo, text in voices.items():
    if vo in files:
        if "[Unintelligible]" in text:
            continue

        with open("{}/{}.txt".format(voice_folder, vo), "w") as f:
            f.write(text)

        audio = sound.fix_audio(BinaryReader.load_file(files[vo]))
        # audio = BinaryReader.load_file(files[vo])
        with open("{}/{}.wav".format(voice_folder, vo), "wb") as f:
            f.write(audio)



import platform
from unittest import TestCase

from pykotor.resource.formats.erf import ERF, ERFBinaryReader, write_erf, read_erf
from pykotor.resource.type import ResourceType

BINARY_TEST_FILE = "../../files/test.erf"
DOES_NOT_EXIST_FILE = "./thisfiledoesnotexist"
CORRUPT_BINARY_TEST_FILE = "../../files/test_corrupted.gff"


class TestERF(TestCase):
    def test_binary_io(self):
        erf = ERFBinaryReader(BINARY_TEST_FILE).load()
        self.validate_io(erf)

        data = bytearray()
        write_erf(erf, data)
        erf = ERFBinaryReader(data).load()
        self.validate_io(erf)

    def validate_io(self, erf: ERF):
        self.assertEqual(len(erf), 3)
        self.assertEqual(erf.get("1", ResourceType.TXT), b'abc')
        self.assertEqual(erf.get("2", ResourceType.TXT), b'def')
        self.assertEqual(erf.get("3", ResourceType.TXT), b'ghi')
    
    def test_read_raises(self):
        if platform.system() == "Windows":
            self.assertRaises(PermissionError, read_erf, ".")
        else:
            self.assertRaises(IsADirectoryError, read_erf, ".")
        self.assertRaises(FileNotFoundError, read_erf, DOES_NOT_EXIST_FILE)
        self.assertRaises(ValueError, read_erf, CORRUPT_BINARY_TEST_FILE)

    def test_write_raises(self):
        if platform.system() == "Windows":
            self.assertRaises(PermissionError, write_erf, ERF(), ".", ResourceType.ERF)
        else:
            self.assertRaises(IsADirectoryError, write_erf, ERF(), ".", ResourceType.ERF)
        self.assertRaises(ValueError, write_erf, ERF(), ".", ResourceType.INVALID)
    


from pykotor.resource.formats.erf.erf_data import ERF, ERFResource, ERFType
from pykotor.resource.formats.erf.io_erf import ERFBinaryReader, ERFBinaryWriter
from pykotor.resource.formats.erf.erf_auto import read_erf, write_erf



from unittest import TestCase

from pykotor.resource.formats.gff import read_gff
from pykotor.resource.generics.utt import construct_utt, dismantle_utt

TEST_FILE = "../../files/test.utt"


class TestUTT(TestCase):
    def test_io(self):
        gff = read_gff(TEST_FILE)
        utt = construct_utt(gff)
        self.validate_io(utt)

        gff = dismantle_utt(utt)
        utt = construct_utt(gff)
        self.validate_io(utt)

    def validate_io(self, utt):
        self.assertEqual("GenericTrigger001", utt.tag)
        self.assertEqual("generictrigge001", utt.resref)
        self.assertEqual(42968, utt.name.stringref)
        self.assertEqual(1, utt.auto_remove_key)
        self.assertEqual(1, utt.faction_id)
        self.assertEqual(1, utt.cursor_id)
        self.assertEqual(3.0, utt.highlight_height)
        self.assertEqual("somekey", utt.key_name)
        self.assertEqual(0, utt.loadscreen_id)
        self.assertEqual(0, utt.portrait_id)
        self.assertEqual(1, utt.type_id)
        self.assertEqual(1, utt.trap_detectable)
        self.assertEqual(10, utt.trap_detect_dc)
        self.assertEqual(1, utt.trap_disarmable)
        self.assertEqual(10, utt.trap_disarm_dc)
        self.assertEqual(1, utt.is_trap)
        self.assertEqual(1, utt.trap_once)
        self.assertEqual(1, utt.trap_type)
        self.assertEqual("ondisarm", utt.on_disarm)
        self.assertEqual("ontraptriggered", utt.on_trap_triggered)
        self.assertEqual("onclick", utt.on_click)
        self.assertEqual("onheartbeat", utt.on_heartbeat)
        self.assertEqual("onenter", utt.on_enter)
        self.assertEqual("onexit", utt.on_exit)
        self.assertEqual("onuserdefined", utt.on_user_defined)
        self.assertEqual(6, utt.palette_id)
        self.assertEqual("comment", utt.comment)





from pykotor.resource.formats.ssf.ssf_data import SSF, SSFSound
from pykotor.resource.formats.ssf.io_ssf import SSFBinaryReader, SSFBinaryWriter
from pykotor.resource.formats.ssf.io_ssf_xml import SSFXMLReader, SSFXMLWriter
from pykotor.resource.formats.ssf.ssf_auto import detect_ssf, read_ssf, write_ssf



from pykotor.common.script import ScriptFunction

from pykotor.common.stream import BinaryReader

nwscript_file = "C:/New folder/nwscript.nss"
file = BinaryReader.load_file(nwscript_file).decode(errors='ignore')
file = file.split("\n")[1666:]

functions = []
function = ScriptFunction("void", "", [], "", "")
for line in file:
    if line.startswith("//"):
        function = ScriptFunction("void", "", [], "", "")
        #routine_id = line[3:line.index(".")]
        print(line)




from unittest import TestCase

from pykotor.extract.capsule import Capsule
from pykotor.resource.type import ResourceType

TEST_FILE_ERF = "../../tests/files/capsule.mod"
TEST_FILE_RIM = "../../tests/files/capsule.rim"


class TestCapsule(TestCase):
    def test_erf_capsule(self):
        erf_capsule = Capsule(TEST_FILE_ERF)

        self.assertEqual(3, len(erf_capsule))

        self.assertTrue(erf_capsule.exists("001ebo", ResourceType.ARE))
        self.assertEqual(4865, len(erf_capsule.resource("001ebo", ResourceType.ARE)))
        self.assertEqual("ARE ", erf_capsule.resource("001ebo", ResourceType.ARE)[:4].decode())

        self.assertTrue(erf_capsule.exists("001ebo", ResourceType.GIT))
        self.assertEqual(42565, len(erf_capsule.resource("001ebo", ResourceType.GIT)))
        self.assertEqual("GIT ", erf_capsule.resource("001ebo", ResourceType.GIT)[:4].decode())

        self.assertTrue(erf_capsule.exists("001ebo", ResourceType.PTH))
        self.assertEqual(19788, len(erf_capsule.resource("001ebo", ResourceType.PTH)))
        self.assertEqual("PTH ", erf_capsule.resource("001ebo", ResourceType.PTH)[:4].decode())

    def test_rim_capsule(self):
        rim_capsule = Capsule(TEST_FILE_RIM)

        self.assertEqual(3, len(rim_capsule))

        self.assertTrue(rim_capsule.exists("m13aa", ResourceType.ARE))
        self.assertEqual(4096, len(rim_capsule.resource("m13aa", ResourceType.ARE)))
        self.assertEqual("ARE ", rim_capsule.resource("m13aa", ResourceType.ARE)[:4].decode())

        self.assertTrue(rim_capsule.exists("m13aa", ResourceType.GIT))
        self.assertEqual(51747, len(rim_capsule.resource("m13aa", ResourceType.GIT)))
        self.assertEqual("GIT ", rim_capsule.resource("m13aa", ResourceType.GIT)[:4].decode())

        self.assertTrue(rim_capsule.exists("module", ResourceType.IFO))
        self.assertEqual(1655, len(rim_capsule.resource("module", ResourceType.IFO)))
        self.assertEqual("IFO ", rim_capsule.resource("module", ResourceType.IFO)[:4].decode())



from __future__ import annotations

from pykotor.common.geometry import Vector3, Vector2
from pykotor.common.language import LocalizedString
from pykotor.common.misc import Game, ResRef
from pykotor.resource.formats.gff import GFF, GFFList, GFFContent, read_gff, write_gff
from pykotor.resource.formats.gff.gff_auto import bytes_gff
from pykotor.resource.type import ResourceType, SOURCE_TYPES, TARGET_TYPES


class IFO:
    """
    Stores module information.

    Attributes:
        mod_id: "Mod_ID" field.
        vo_id: "Mod_VO_ID" field.
        mod_name: "Mod_Name" field.
        tag: "Mod_Tag" field.
        identifier: "Mod_Entry_Area" field.
        entry_position.x: "Mod_Entry_X" field.
        entry_position.y: "Mod_Entry_Y" field.
        entry_position.z: "Mod_Entry_Z" field.
        entry_direction.x: "Mod_Entry_Dir_X" field.
        entry_direction.y: "Mod_Entry_Dir_Y" field.
        on_heartbeat: "Mod_OnHeartbeat" field.
        on_load: "Mod_OnModLoad" field.
        on_start: "Mod_OnModStart" field.
        on_enter: "Mod_OnClientEntr" field.
        on_leave: "Mod_OnClientLeav" field.
        on_activate_item: "Mod_OnActvtItem" field.
        on_acquire_item: "Mod_OnAcquirItem" field.
        on_user_defined: "Mod_OnUsrDefined" field.
        on_unacquire_item: "Mod_OnUnAqreItem" field.
        on_player_death: "Mod_OnPlrDeath" field.
        on_player_dying: "Mod_OnPlrDying" field.
        on_player_levelup: "Mod_OnPlrLvlUp" field.
        on_player_respawn: "Mod_OnSpawnBtnDn" field.
        expansion_id: "Expansion_Pack" field.

        hak: "Mod_Hak" field. Not used by the game engine.
        description: "Mod_Description" field. Not used by the game engine.
        on_player_rest: "Mod_OnPlrRest" field. Not used by the game engine.
        dawn_hour: "Mod_DawnHour" field. Not used by the game engine.
        dusk_hour: "Mod_DuskHour" field. Not used by the game engine.
        time_scale: "Mod_MinPerHour" field. Not used by the game engine.
        start_month: "Mod_StartMonth" field. Not used by the game engine.
        start_day: "Mod_StartDay" field. Not used by the game engine.
        start_hour: "Mod_StartHour" field. Not used by the game engine.
        start_year: "Mod_StartYear" field. Not used by the game engine.
        xp_scale: "Mod_XPScale" field. Not used by the game engine.
        start_movie: "Mod_StartMovie" field. Not used by the game engine.
        creator_id: "Mod_Creator_ID" field. Not used by the game engine.
        version: "Mod_Version" field. Not used by the game engine.
    """

    BINARY_TYPE = ResourceType.IFO

    def __init__(
            self
    ):
        self.mod_id: bytes = bytes()
        self.mod_name: LocalizedString = LocalizedString.from_invalid()
        self.area_name: ResRef = ResRef.from_blank()

        self.identifier: ResRef = ResRef.from_blank()
        self.entry_direction: float = 0.0
        self.entry_position: Vector3 = Vector3.from_null()
        self.tag: str = ""
        self.vo_id: str = ""

        self.on_acquire_item: ResRef = ResRef.from_blank()
        self.on_activate_item: ResRef = ResRef.from_blank()
        self.on_enter: ResRef = ResRef.from_blank()
        self.on_leave: ResRef = ResRef.from_blank()
        self.on_heartbeat: ResRef = ResRef.from_blank()
        self.on_load: ResRef = ResRef.from_blank()
        self.on_start: ResRef = ResRef.from_blank()
        self.on_player_death: ResRef = ResRef.from_blank()
        self.on_player_dying: ResRef = ResRef.from_blank()
        self.on_player_levelup: ResRef = ResRef.from_blank()
        self.on_player_respawn: ResRef = ResRef.from_blank()
        self.on_unacquire_item: ResRef = ResRef.from_blank()
        self.on_user_defined: ResRef = ResRef.from_blank()

        # Deprecated:
        self.creator_id: int = 0
        self.version: int = 0
        self.expansion_id: int = 0
        self.hak: str = ""
        self.description: LocalizedString = LocalizedString.from_invalid()
        self.dawn_hour: int = 0
        self.dusk_hour: int = 0
        self.time_scale: int = 0
        self.start_month: int = 0
        self.start_day: int = 0
        self.start_hour: int = 0
        self.start_year: int = 0
        self.xp_scale: int = 0
        self.on_player_rest: ResRef = ResRef.from_blank()
        self.start_movie: ResRef = ResRef.from_blank()


def construct_ifo(
        gff: GFF
) -> IFO:
    ifo = IFO()

    root = gff.root
    ifo.mod_id = root.acquire("Mod_ID", bytes())
    ifo.vo_id = root.acquire("Mod_VO_ID", "")
    ifo.mod_name = root.acquire("Mod_Name", LocalizedString.from_invalid())
    ifo.tag = root.acquire("Mod_Tag", "")
    ifo.identifier = root.acquire("Mod_Entry_Area", ResRef.from_blank())
    ifo.entry_position.x = root.acquire("Mod_Entry_X", 0.0)
    ifo.entry_position.y = root.acquire("Mod_Entry_Y", 0.0)
    ifo.entry_position.z = root.acquire("Mod_Entry_Z", 0.0)
    ifo.on_heartbeat = root.acquire("Mod_OnHeartbeat", ResRef.from_blank())
    ifo.on_load = root.acquire("Mod_OnModLoad", ResRef.from_blank())
    ifo.on_start = root.acquire("Mod_OnModStart", ResRef.from_blank())
    ifo.on_enter = root.acquire("Mod_OnClientEntr", ResRef.from_blank())
    ifo.on_leave = root.acquire("Mod_OnClientLeav", ResRef.from_blank())
    ifo.on_activate_item = root.acquire("Mod_OnActvtItem", ResRef.from_blank())
    ifo.on_acquire_item = root.acquire("Mod_OnAcquirItem", ResRef.from_blank())
    ifo.on_user_defined = root.acquire("Mod_OnUsrDefined", ResRef.from_blank())
    ifo.on_unacquire_item = root.acquire("Mod_OnUnAqreItem", ResRef.from_blank())
    ifo.on_player_death = root.acquire("Mod_OnPlrDeath", ResRef.from_blank())
    ifo.on_player_dying = root.acquire("Mod_OnPlrDying", ResRef.from_blank())
    ifo.on_player_levelup = root.acquire("Mod_OnPlrLvlUp", ResRef.from_blank())
    ifo.on_player_respawn = root.acquire("Mod_OnSpawnBtnDn", ResRef.from_blank())
    ifo.expansion_id = root.acquire("Expansion_Pack", 0)
    ifo.hak = root.acquire("Mod_Hak", "")
    ifo.description = root.acquire("Mod_Description", LocalizedString.from_invalid())
    ifo.on_player_rest = root.acquire("Mod_OnPlrRest", ResRef.from_blank())
    ifo.dawn_hour = root.acquire("Mod_DawnHour", 0)
    ifo.dusk_hour = root.acquire("Mod_DuskHour", 0)
    ifo.time_scale = root.acquire("Mod_MinPerHour", 0)
    ifo.start_month = root.acquire("Mod_StartMonth", 0)
    ifo.start_day = root.acquire("Mod_StartDay", 0)
    ifo.start_hour = root.acquire("Mod_StartHour", 0)
    ifo.start_year = root.acquire("Mod_StartYear", 0)
    ifo.xp_scale = root.acquire("Mod_XPScale", 0)
    ifo.start_movie = root.acquire("Mod_StartMovie", ResRef.from_blank())
    ifo.creator_id = root.acquire("Mod_Creator_ID", 0)
    ifo.version = root.acquire("Mod_Version", 0)

    dir_x = root.acquire("Mod_Entry_Dir_X", 0.0)
    dir_y = root.acquire("Mod_Entry_Dir_Y", 0.0)
    ifo.entry_direction = Vector2(dir_x, dir_y).angle()

    ifo.area_name = root.acquire("Mod_Area_list", GFFList()).at(0).acquire("Area_Name", ResRef.from_blank())

    return ifo


def dismantle_ifo(
        ifo: IFO,
        game: Game = Game.K2,
        *,
        use_deprecated: bool = True
) -> GFF:
    gff = GFF(GFFContent.IFO)

    root = gff.root
    root.set_binary("Mod_ID", ifo.mod_id)
    root.set_string("Mod_VO_ID", ifo.vo_id)
    root.set_locstring("Mod_Name", ifo.mod_name)
    root.set_string("Mod_Tag", ifo.tag)
    root.set_uint8("Mod_IsSaveGame", 0)
    root.set_resref("Mod_Entry_Area", ifo.identifier)
    root.set_single("Mod_Entry_X", ifo.entry_position.x)
    root.set_single("Mod_Entry_Y", ifo.entry_position.y)
    root.set_single("Mod_Entry_Z", ifo.entry_position.z)
    root.set_resref("Mod_OnHeartbeat", ifo.on_heartbeat)
    root.set_resref("Mod_OnModLoad", ifo.on_load)
    root.set_resref("Mod_OnModStart", ifo.on_start)
    root.set_resref("Mod_OnClientEntr", ifo.on_enter)
    root.set_resref("Mod_OnClientLeav", ifo.on_leave)
    root.set_resref("Mod_OnActvtItem", ifo.on_activate_item)
    root.set_resref("Mod_OnAcquirItem", ifo.on_acquire_item)
    root.set_resref("Mod_OnUsrDefined", ifo.on_user_defined)
    root.set_resref("Mod_OnUnAqreItem", ifo.on_unacquire_item)
    root.set_resref("Mod_OnPlrDeath", ifo.on_player_death)
    root.set_resref("Mod_OnPlrDying", ifo.on_player_dying)
    root.set_resref("Mod_OnPlrLvlUp", ifo.on_player_levelup)
    root.set_resref("Mod_OnSpawnBtnDn", ifo.on_player_respawn)

    entry_direction = Vector2.from_angle(ifo.entry_direction)
    root.set_single("Mod_Entry_Dir_X", entry_direction.x)
    root.set_single("Mod_Entry_Dir_Y", entry_direction.y)

    root.set_list("Mod_Area_list", GFFList()).add(6).set_resref("Area_Name", ifo.area_name)

    if use_deprecated:
        root.set_uint16("Expansion_Pack", ifo.expansion_id)
        root.set_string("Mod_Hak", ifo.hak)
        root.set_locstring("Mod_Description", ifo.description)
        root.set_resref("Mod_OnPlrRest", ifo.on_player_rest)
        root.set_uint8("Mod_DawnHour", ifo.dawn_hour)
        root.set_uint8("Mod_DuskHour", ifo.dusk_hour)
        root.set_uint8("Mod_MinPerHour", ifo.time_scale)
        root.set_uint8("Mod_StartMonth", ifo.start_month)
        root.set_uint8("Mod_StartDay", ifo.start_day)
        root.set_uint8("Mod_StartHour", ifo.start_hour)
        root.set_uint32("Mod_StartYear", ifo.start_year)
        root.set_uint8("Mod_XPScale", ifo.xp_scale)
        root.set_resref("Mod_StartMovie", ifo.start_movie)
        root.set_int32("Mod_Creator_ID", ifo.creator_id)
        root.set_uint32("Mod_Version", ifo.version)

    return gff


def read_ifo(
        source: SOURCE_TYPES,
        offset: int = 0,
        size: int = None
) -> IFO:
    gff = read_gff(source, offset, size)
    ifo = construct_ifo(gff)
    return ifo


def write_ifo(
        ifo: IFO,
        target: TARGET_TYPES,
        game: Game = Game.K2,
        file_format: ResourceType = ResourceType.GFF,
        *,
        use_deprecated: bool = True
) -> None:
    gff = dismantle_ifo(ifo, game, use_deprecated=use_deprecated)
    write_gff(gff, target, file_format)


def bytes_ifo(
        ifo: IFO,
        game: Game = Game.K2,
        file_format: ResourceType = ResourceType.GFF,
        *,
        use_deprecated: bool = True
) -> bytes:
    gff = dismantle_ifo(ifo, game, use_deprecated=use_deprecated)
    return bytes_gff(gff, file_format)



from pykotor.resource.formats.ltr.io_ltr import LTRBinaryReader, LTRBinaryWriter
from pykotor.resource.formats.ltr.ltr_data import LTR
from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES, ResourceType


def read_ltr(
        source: SOURCE_TYPES,
        offset: int = 0,
        size: int = None
) -> LTR:
    """
    Returns an LTR instance from the source.

    Args:
        source: The source of the data.
        offset: The byte offset of the file inside the data.
        size: Number of bytes to allowed to read from the stream. If not specified, uses the whole stream.

    Raises:
        ValueError: If the file was corrupted or in an unsupported format.

    Returns:
        An LTR instance.
    """
    try:
        return LTRBinaryReader(source, offset, size).load()
    except (IOError, ValueError):
        raise ValueError("Tried to load an unsupported or corrupted LTR file.")


def write_ltr(
        ltr: LTR,
        target: TARGET_TYPES,
        file_format: ResourceType = ResourceType.LTR
) -> None:
    """
    Writes the LTR data to the target location with the specified format (LTR only).

    Args:
        ltr: The LTR file being written.
        target: The location to write the data to.
        file_format: The file format.

    Raises:
        ValueError: If an unsupported file format was given.
    """
    if file_format == ResourceType.LTR:
        LTRBinaryWriter(ltr, target).write()
    else:
        raise ValueError("Unsupported format specified; use LTR.")


def bytes_ltr(
        ltr: LTR,
        file_format: ResourceType = ResourceType.LTR
) -> bytes:
    """
    Returns the LTR data in the specified format (LTR only) as a bytes object.

    This is a convience method that wraps the write_ltr() method.

    Args:
        ltr: The target LTR object.
        file_format: The file format.

    Raises:
        ValueError: If an unsupported file format was given.

    Returns:
        The LTR data.
    """
    data = bytearray()
    write_ltr(ltr, data, file_format)
    return data



from unittest import TestCase

from pykotor.common.misc import Color
from pykotor.resource.formats.gff import read_gff
from pykotor.resource.generics.git import construct_git, dismantle_git

TEST_FILE = "../../files/test.git"


class TestGIT(TestCase):
    def test_io(self):
        gff = read_gff(TEST_FILE)
        git = construct_git(gff)
        self.validate_io(git)

        gff = dismantle_git(git)
        git = construct_git(gff)
        self.validate_io(git)

    def validate_io(self, git):
        self.assertEqual(127, git.ambient_volume)
        self.assertEqual(17, git.ambient_sound_id)
        self.assertEqual(1, git.env_audio)
        self.assertEqual(41, git.music_battle_id)
        self.assertEqual(15, git.music_standard_id)
        self.assertEqual(20000, git.music_delay)

        self.assertEqual(1, git.cameras[0].camera_id)
        self.assertEqual(55, git.cameras[0].fov)
        self.assertEqual(3.0, git.cameras[0].height)
        self.assertEqual(0.0, git.cameras[0].mic_range)
        self.assertAlmostEqual(69.699, git.cameras[0].pitch, 2)
        self.assertAlmostEqual(0.971, git.cameras[0].orientation.x, 2)
        self.assertAlmostEqual(0.000, git.cameras[0].orientation.y, 2)
        self.assertAlmostEqual(0.000, git.cameras[0].orientation.z, 2)
        self.assertAlmostEqual(0.235, git.cameras[0].orientation.w, 2)
        self.assertAlmostEqual(-57.167, git.cameras[0].position.x, 2)
        self.assertAlmostEqual(-28.255, git.cameras[0].position.y, 2)
        self.assertAlmostEqual(0.000, git.cameras[0].position.z, 2)

        self.assertEqual("c_ithorian001", git.creatures[0].resref)
        self.assertAlmostEqual(-41.238, git.creatures[0].position.x, 2)
        self.assertAlmostEqual(-53.214, git.creatures[0].position.y, 2)
        self.assertAlmostEqual(0.000, git.creatures[0].position.z, 2)
        self.assertAlmostEqual(146.249, git.creatures[0].bearing, 2)

        self.assertAlmostEqual(1.0, git.doors[0].bearing, 2)
        self.assertAlmostEqual(-43.763, git.doors[0].position.x, 2)
        self.assertAlmostEqual(-20.143, git.doors[0].position.y, 2)
        self.assertAlmostEqual(1.000, git.doors[0].position.z, 2)
        self.assertEqual("linkedto", git.doors[0].linked_to)
        self.assertEqual(1, git.doors[0].linked_to_flags.value)
        self.assertEqual("resref", git.doors[0].linked_to_module)
        self.assertEqual("Ithorian", git.doors[0].tag)
        self.assertEqual("sw_door_taris007", git.doors[0].resref)
        self.assertEqual(13, git.doors[0].transition_destination.stringref)
        self.assertEqual(Color.from_bgr_integer(10197915), git.doors[0].tweak_color)

        self.assertEqual("mercenariesentry", git.encounters[0].resref)
        self.assertAlmostEqual(-41.319, git.encounters[0].position.x, 2)
        self.assertAlmostEqual(-19.222, git.encounters[0].position.y, 2)
        self.assertAlmostEqual(1.000, git.encounters[0].position.z, 2)
        self.assertAlmostEqual(-5.890, git.encounters[0].geometry[0].x, 2)
        self.assertAlmostEqual(3.072, git.encounters[0].geometry[0].y, 2)
        self.assertAlmostEqual(0.025, git.encounters[0].geometry[0].z, 2)
        self.assertAlmostEqual(-48.936, git.encounters[0].spawn_points[0].position.x, 2)
        self.assertAlmostEqual(-29.831, git.encounters[0].spawn_points[0].position.y, 2)
        self.assertAlmostEqual(1.000, git.encounters[0].spawn_points[0].position.z, 2)
        self.assertAlmostEqual(0.196, git.encounters[0].spawn_points[0].orientation, 2)

        self.assertEqual("k_trans_abort", git.placeables[0].resref)
        self.assertAlmostEqual(1.0, git.placeables[0].bearing, 2)
        self.assertAlmostEqual(-33.268, git.placeables[0].position.x, 2)
        self.assertAlmostEqual(-15.299, git.placeables[0].position.y, 2)
        self.assertAlmostEqual(9.536, git.placeables[0].position.z, 2)
        self.assertEqual(Color.from_bgr_integer(10197915), git.placeables[0].tweak_color)

        self.assertEqual("computerpanne001", git.sounds[0].resref)
        self.assertAlmostEqual(-78.538, git.sounds[0].position.x, 2)
        self.assertAlmostEqual(13.498, git.sounds[0].position.y, 2)
        self.assertAlmostEqual(2.000, git.sounds[0].position.z, 2)

        self.assertEqual("m_chano", git.stores[0].resref)
        self.assertAlmostEqual(106.230, git.stores[0].position.x, 2)
        self.assertAlmostEqual(-16.590, git.stores[0].position.y, 2)
        self.assertAlmostEqual(0.063, git.stores[0].position.z, 2)
        self.assertAlmostEqual(90.000, git.stores[0].bearing, 2)

        self.assertEqual("newgeneric001", git.triggers[0].resref)
        self.assertAlmostEqual(-29.903, git.triggers[0].position.x, 2)
        self.assertAlmostEqual(-11.463, git.triggers[0].position.y, 2)
        self.assertAlmostEqual(-2.384, git.triggers[0].position.z, 2)
        self.assertEqual("from_204TEL", git.triggers[0].linked_to)
        self.assertEqual(2, git.triggers[0].linked_to_flags.value)
        self.assertEqual("203tel", git.triggers[0].linked_to_module)
        self.assertEqual("to_203TEL", git.triggers[0].tag)
        self.assertEqual(104245, git.triggers[0].transition_destination.stringref)
        self.assertAlmostEqual(-7.433, git.triggers[0].geometry[0].x, 2)
        self.assertAlmostEqual(1.283, git.triggers[0].geometry[0].y, 2)
        self.assertAlmostEqual(0.025, git.triggers[0].geometry[0].z, 2)

        self.assertEqual("wp_transabort", git.waypoints[0].resref)
        self.assertEqual("wp_transabort", git.waypoints[0].tag)
        self.assertEqual(135283, git.waypoints[0].name.stringref)
        self.assertTrue(git.waypoints[0].map_note_enabled)
        self.assertEqual(123, git.waypoints[0].map_note.stringref)
        self.assertAlmostEqual(-33.620, git.waypoints[0].position.x, 2)
        self.assertAlmostEqual(-16.065, git.waypoints[0].position.y, 2)
        self.assertAlmostEqual(1.0, git.waypoints[0].position.z, 2)
        self.assertAlmostEqual(90.000, git.waypoints[0].bearing, 2)



from unittest import TestCase

from pykotor.resource.formats.gff import read_gff
from pykotor.resource.generics.ute import construct_ute, dismantle_ute

TEST_FILE = "../../files/test.ute"


class TestUTE(TestCase):
    def test_io(self):
        gff = read_gff(TEST_FILE)
        ute = construct_ute(gff)
        self.validate_io(ute)

        gff = dismantle_ute(ute)
        ute = construct_ute(gff)
        self.validate_io(ute)

    def validate_io(self, ute):
        self.assertEqual("G_KATAARNGROUP01", ute.tag)
        self.assertEqual(31918, ute.name.stringref)
        self.assertEqual("g_kataarngroup01", ute.resref)
        self.assertEqual(1, ute.active)
        self.assertEqual(1, ute.unused_difficulty)
        self.assertEqual(2, ute.difficulty_id)
        self.assertEqual(1, ute.faction_id)
        self.assertEqual(6, ute.max_creatures)
        self.assertEqual(1, ute.player_only)
        self.assertEqual(3, ute.rec_creatures)
        self.assertEqual(1, ute.reset)
        self.assertEqual(60, ute.reset_time)
        self.assertEqual(1, ute.respawns)
        self.assertEqual(1, ute.single_shot)
        self.assertEqual("onentered", ute.on_entered)
        self.assertEqual("onexit", ute.on_exit)
        self.assertEqual("onexhausted", ute.on_exhausted)
        self.assertEqual("onheartbeat", ute.on_heartbeat)
        self.assertEqual("onuserdefined", ute.on_user_defined)
        self.assertEqual(7, ute.palette_id)
        self.assertEqual("Kashyyyk", ute.comment)

        self.assertEqual(2, len(ute.creatures))
        self.assertEqual(74, ute.creatures[1].appearance_id)
        self.assertEqual(8.0, ute.creatures[1].challenge_rating)
        self.assertEqual("g_kataarn02", ute.creatures[1].resref)
        self.assertEqual(1, ute.creatures[1].guaranteed_count)
        self.assertTrue(ute.creatures[1].single_spawn)




from unittest import TestCase

from pykotor.resource.formats.gff import read_gff
from pykotor.resource.generics.utm import construct_utm, dismantle_utm

TEST_FILE = "../../files/test.utm"


class TestUTM(TestCase):
    def test_io(self):
        gff = read_gff(TEST_FILE)
        utm = construct_utm(gff)
        self.validate_io(utm)

        gff = dismantle_utm(utm)
        utm = construct_utm(gff)
        self.validate_io(utm)

    def validate_io(self, utm):
        self.assertEqual("dan_droid", utm.resref)
        self.assertEqual(33399, utm.name.stringref)
        self.assertEqual("dan_droid", utm.tag)
        self.assertEqual(100, utm.mark_up)
        self.assertEqual(25, utm.mark_down)
        self.assertEqual("onopenstore", utm.on_open)
        self.assertEqual("comment", utm.comment)
        self.assertEqual(5, utm.id)
        self.assertTrue(utm.can_buy)
        self.assertTrue(utm.can_sell)

        self.assertEqual(2, len(utm.inventory))
        self.assertFalse(utm.inventory[0].infinite)
        self.assertTrue(utm.inventory[1].infinite)
        self.assertEqual("g_i_drdltplat002", utm.inventory[1].resref)




from unittest import TestCase

from pykotor.common.stream import BinaryReader

from pykotor.resource.formats.lip import LIP
from pykotor.resource.formats.ncs import NCSBinaryReader, NCS, NCSBinaryWriter
from pykotor.resource.formats.ncs.ncs_auto import bytes_ncs, write_ncs, read_ncs

BINARY_TEST_FILE = "../../files/test.ncs"


class TestNCS(TestCase):
    def test_binary_io(self):
        ncs = NCSBinaryReader(BINARY_TEST_FILE).load()
        self.validate_io(ncs)

        write_ncs(ncs, r"C:\Users\hugin\Desktop\ext\output.ncs")
        data = bytes_ncs(ncs)
        ncs = read_ncs(data)
        self.validate_io(ncs)

    def validate_io(self, ncs: NCS):
        self.assertEqual(8, len(ncs.instructions))

        self.assertEqual(BinaryReader.load_file(BINARY_TEST_FILE), bytes_ncs(ncs))



from unittest import TestCase

from pykotor.resource.formats.gff import read_gff
from pykotor.resource.generics.are import construct_are, dismantle_are

TEST_FILE = "../../files/test.are"


class TestARE(TestCase):
    def test_io(self):
        gff = read_gff(TEST_FILE)
        are = construct_are(gff)
        self.validate_io(are)

        gff = dismantle_are(are)
        are = construct_are(gff)
        self.validate_io(are)

    def validate_io(self, are):
        self.assertEqual(0, are.unused_id)
        self.assertEqual(0, are.creator_id)
        self.assertEqual("Untitled", are.tag)
        self.assertEqual(75101, are.name.stringref)
        self.assertEqual("comments", are.comment)
        self.assertEqual(0, are.flags)
        self.assertEqual(0, are.mod_spot_check)
        self.assertEqual(0, are.mod_listen_check)
        self.assertEqual(0.20000000298023224, are.alpha_test)
        self.assertEqual(1, are.camera_style)
        self.assertEqual("defaultenvmap", are.default_envmap)
        self.assertEqual("grasstexture", are.grass_texture)
        self.assertEqual(1.0, are.grass_density)
        self.assertEqual(1.0, are.grass_size)
        self.assertEqual(0.25, are.grass_prob_ll)
        self.assertEqual(0.25, are.grass_prob_lr)
        self.assertEqual(0.25, are.grass_prob_ul)
        self.assertEqual(0.25, are.grass_prob_ur)
        self.assertEqual(0, are.moon_ambient)
        self.assertEqual(0, are.moon_diffuse)
        self.assertEqual(0, are.moon_fog)
        self.assertEqual(99.0, are.moon_fog_near)
        self.assertEqual(100.0, are.moon_fog_far)
        self.assertEqual(0, are.moon_fog_color)
        self.assertEqual(0, are.moon_shadows)
        self.assertEqual(1, are.fog_enabled)
        self.assertEqual(99.0, are.fog_near)
        self.assertEqual(100.0, are.fog_far)
        self.assertEqual(1, are.shadows)
        self.assertEqual(0, are.is_night)
        self.assertEqual(0, are.lighting_scheme)
        self.assertEqual(205, are.shadow_opacity)
        self.assertEqual(0, are.day_night)
        self.assertEqual(99, are.chance_rain)
        self.assertEqual(99, are.chance_snow)
        self.assertEqual(99, are.chance_lightning)
        self.assertEqual(1, are.wind_power)
        self.assertEqual(0, are.loadscreen_id)
        self.assertEqual(3, are.player_vs_player)
        self.assertEqual(0, are.no_rest)
        self.assertEqual(0, are.no_hang_back)
        self.assertEqual(0, are.player_only)
        self.assertEqual(1, are.unescapable)
        self.assertEqual(1, are.disable_transit)
        self.assertEqual(1, are.stealth_xp)
        self.assertEqual(25, are.stealth_xp_loss)
        self.assertEqual(25, are.stealth_xp_max)
        self.assertEqual(1, are.dirty_size_1)
        self.assertEqual(1, are.dirty_formula_1)
        self.assertEqual(1, are.dirty_func_1)
        self.assertEqual(1, are.dirty_size_2)
        self.assertEqual(1, are.dirty_formula_2)
        self.assertEqual(1, are.dirty_func_2)
        self.assertEqual(1, are.dirty_size_3)
        self.assertEqual(1, are.dirty_formula_3)
        self.assertEqual(1, are.dirty_func_3)
        self.assertEqual("k_on_enter", are.on_enter)
        self.assertEqual("onexit", are.on_exit)
        self.assertEqual("onheartbeat", are.on_heartbeat)
        self.assertEqual("onuserdefined", are.on_user_defined)

        self.assertEqual(88, are.version)
        self.assertEqual(16777215, are.grass_ambient.bgr_integer())
        self.assertEqual(16777215, are.grass_diffuse.bgr_integer())
        self.assertEqual(16777215, are.sun_ambient.bgr_integer())
        self.assertEqual(16777215, are.sun_diffuse.bgr_integer())
        self.assertEqual(16777215, are.fog_color.bgr_integer())
        self.assertEqual(16777215, are.dynamic_light.bgr_integer())
        self.assertEqual(123, are.dirty_argb_1.bgr_integer())
        self.assertEqual(1234, are.dirty_argb_2.bgr_integer())
        self.assertEqual(12345, are.dirty_argb_3.bgr_integer())




from __future__ import annotations

from typing import Optional, List

from pykotor.resource.formats.vis import VIS
from pykotor.resource.type import SOURCE_TYPES, TARGET_TYPES, ResourceReader, ResourceWriter, autoclose


class VISAsciiReader(ResourceReader):
    def __init__(
            self,
            source: SOURCE_TYPES,
            offset: int = 0,
            size: int = 0
    ):
        super().__init__(source, offset, size)
        self._vis: Optional[VIS] = None
        self._lines: List[str] = []

    @autoclose
    def load(
            self,
            auto_close: bool = True
    ) -> VIS:
        self._vis = VIS()
        self._lines = self._reader.read_string(self._reader.size()).splitlines()

        pairs = []

        iterator = iter(self._lines)
        for line in iterator:
            tokens = line.split()

            when_inside = tokens[0]
            self._vis.add_room(when_inside)

            count = int(tokens[1])
            for i in range(count):
                show = next(iterator).split()[0]
                pairs.append((when_inside, show))

        for when_inside, show in pairs:
            if when_inside not in self._vis.all_rooms():
                self._vis.add_room(when_inside)
            if show not in self._vis.all_rooms():
                self._vis.add_room(show)
            self._vis.set_visible(when_inside, show, True)

        return self._vis


class VISAsciiWriter(ResourceWriter):
    def __init__(
            self,
            vis: VIS,
            target: TARGET_TYPES
    ):
        super().__init__(target)
        self._vis: VIS = vis

    @autoclose
    def write(
            self,
            auto_close: bool = True
    ) -> None:
        for observer, observed in self._vis:
            self._writer.write_string("{} {}\r\n".format(observer, str(len(observed))))
            for room in observed:
                self._writer.write_string("  {}\r\n".format(room))



import struct

from pykotor.common.stream import BinaryReader
from pykotor.resource.formats.tpc import TPC, TPCTextureFormat
from pykotor.resource.type import ResourceWriter, TARGET_TYPES, autoclose


class TPCBMPWriter(ResourceWriter):
    def __init__(
            self,
            tpc: TPC,
            target: TARGET_TYPES
    ):
        super().__init__(target)
        self._tpc = tpc

    @autoclose
    def write(
            self,
            auto_close: bool = True
    ) -> None:
        width, height, data = self._tpc.convert(TPCTextureFormat.RGB, 0)
        file_size = 14 + 40 + (width * height * 3)

        # Header
        self._writer.write_string("BM")
        self._writer.write_uint32(file_size)
        self._writer.write_uint32(0)
        self._writer.write_uint32(54)

        # InfoHeader
        self._writer.write_uint32(40)
        self._writer.write_uint32(width)
        self._writer.write_uint32(height)
        self._writer.write_uint16(1)
        self._writer.write_uint16(24)
        self._writer.write_uint32(0)
        self._writer.write_uint32(0)
        self._writer.write_uint32(1)
        self._writer.write_uint32(1)
        self._writer.write_uint32(0)  # colors used
        self._writer.write_uint32(0)

        pixel_reader = BinaryReader.from_bytes(data)
        temp_pixels = []
        for i in range(len(data) // 3):
            r = pixel_reader.read_uint8()
            g = pixel_reader.read_uint8()
            b = pixel_reader.read_uint8()
            temp_pixels.append([b, g, r])

        for i in range(len(temp_pixels)):
            x = i % width
            y = height - (i // width) - 1
            index = x + width * y
            self._writer.write_bytes(struct.pack('BBB', *temp_pixels[index]))



import pathlib
from setuptools import setup, find_namespace_packages

HERE = pathlib.Path(__file__).parent

NAME = "PyKotor"
VERSION = "1.6.0"
AUTHOR = "Nicholas Hugi"
DESCRIPTION = "Read, modify and write files used by KotOR's game engine."
PACKAGES = find_namespace_packages(exclude=["tests", "docs"])
URL = "https://github.com/NickHugi/PyKotor"

README = (HERE / "README.md").read_text()
REQUIREMENTS = (HERE / "requirements.txt").read_text()

setup(
    name=NAME,
    version=VERSION,
    author=AUTHOR,
    description=DESCRIPTION,
    install_requires=REQUIREMENTS,
    long_description=README,
    packages=PACKAGES,
    url=URL,
    long_description_content_type='text/markdown'
)



from __future__ import annotations

import os
from contextlib import suppress
from copy import copy
from enum import IntEnum
from typing import Dict, List, Optional, NamedTuple

from pykotor.common.language import Language, Gender, LocalizedString
from pykotor.common.misc import CaseInsensitiveDict
from pykotor.common.stream import BinaryReader
from pykotor.extract.capsule import Capsule
from pykotor.extract.chitin import Chitin
from pykotor.extract.file import FileResource, ResourceResult, LocationResult, ResourceIdentifier
from pykotor.extract.talktable import TalkTable
from pykotor.resource.formats.gff import read_gff
from pykotor.resource.formats.tpc import TPC, read_tpc
from pykotor.resource.type import ResourceType
from pykotor.tools import sound


class SearchLocation(IntEnum):
    OVERRIDE = 0
    """Resources in the installation's override directory and any nested subfolders."""

    MODULES = 1
    """Encapsulated resources in the installation's 'modules' directory."""

    CHITIN = 2
    """Encapsulated resources linked to the installation's 'chitin.key' file."""

    TEXTURES_TPA = 3
    """Encapsulated resources in the installation's 'TexturePacks/swpc_tex_tpa.erf' file."""

    TEXTURES_TPB = 4
    """Encapsulated resources in the installation's 'TexturePacks/swpc_tex_tpb.erf' file."""

    TEXTURES_TPC = 5
    """Encapsulated resources in the installation's 'TexturePacks/swpc_tex_tpc.erf' file."""

    TEXTURES_GUI = 6
    """Encapsulated resources in the installation's 'TexturePacks/swpc_tex_gui.erf' file."""

    MUSIC = 7
    """Resource files in the installation's 'StreamMusic' directory and any nested subfolders."""

    SOUND = 8
    """Resource files in the installation's 'StreamSounds' directory and any nested subfolders."""

    VOICE = 9
    """Resource files in the installation's 'StreamVoice' or 'StreamWaves' directory and any nested subfolders."""

    LIPS = 10
    """Encapsulated resources in the installation's 'lips' directory."""

    RIMS = 11
    """Encapsulated resources in the installation's 'rims' directory."""

    CUSTOM_MODULES = 12
    """Encapsulated resources stored in the capsules specified in method parameters."""

    CUSTOM_FOLDERS = 13
    """Resource files stored in the folders specified in the method parameters."""


class ItemTuple(NamedTuple):
    resname: str
    name: str
    filepath: str


class Installation:
    """
    Installation provides a centralized location for loading resources stored in the game through its
    various folders and formats.
    """
    TEXTURES_TYPES = [ResourceType.TPC, ResourceType.TGA, ResourceType.DDS]

    def __init__(
            self,
            path: str
    ):
        self._path: str = path.replace('\\', '/')
        if not self._path.endswith('/'): self._path += '/'

        self._talktable: Optional[TalkTable] = TalkTable(self._path + "dialog.tlk")

        self._chitin: List[FileResource] = []
        self._modules: Dict[str, List[FileResource]] = {}
        self._lips: Dict[str, List[FileResource]] = {}
        self._texturepacks: Dict[str, List[FileResource]] = {}
        self._override: Dict[str, List[FileResource]] = {}
        self._streammusic: List[FileResource] = []
        self._streamsounds: List[FileResource] = []
        self._streamvoices: List[FileResource] = []
        self._rims: Dict[str, List[FileResource]] = {}

        self.load_modules()
        self.load_override()
        self.load_lips()
        self.load_textures()
        self.load_chitin()
        self.load_streammusic()
        self.load_streamsounds()
        self.load_streamvoices()
        self.load_rims()

    # region Get Paths
    def path(
            self
    ) -> str:
        """
        Returns the path to root folder of the Installation.

        Returns:
            The path to the root folder.
        """
        return self._path.replace("\\", "/")

    def module_path(
            self
    ) -> str:
        """
        Returns the path to modules folder of the Installation. This method maintains the case of the foldername.

        Returns:
            The path to the modules folder.
        """
        module_path = self._path
        for folder in os.listdir(self._path):
            if os.path.isdir(module_path + folder) and folder.lower() == "modules":
                module_path += folder + "/"
        if module_path == self._path:
            raise ValueError("Could not find modules folder in '{}'.".format(self._path))
        return module_path.replace("\\", "/")

    def override_path(
            self
    ) -> str:
        """
        Returns the path to override folder of the Installation. This method maintains the case of the foldername.

        Returns:
            The path to the override folder.
        """
        override_path = self._path
        for folder in os.listdir(self._path):
            if os.path.isdir(override_path + folder) and folder.lower() == "override":
                override_path += folder + "/"
        if override_path == self._path:
            raise ValueError("Could not find override folder in '{}'.".format(self._path))
        return override_path.replace("\\", "/")

    def lips_path(
            self
    ) -> str:
        """
        Returns the path to lips folder of the Installation. This method maintains the case of the foldername.

        Returns:
            The path to the lips folder.
        """
        lips_path = self._path
        for folder in os.listdir(self._path):
            if os.path.isdir(lips_path + folder) and folder.lower() == "lips":
                lips_path += folder + "/"
        if lips_path == self._path:
            raise ValueError("Could not find modules folder in '{}'.".format(self._path))
        return lips_path.replace("\\", "/")

    def texturepacks_path(
            self
    ) -> str:
        """
        Returns the path to texturepacks folder of the Installation. This method maintains the case of the foldername.

        Returns:
            The path to the texturepacks folder.
        """
        texturepacks_path = self._path
        for folder in os.listdir(self._path):
            if os.path.isdir(texturepacks_path + folder) and folder.lower() == "texturepacks":
                texturepacks_path += folder + "/"
        if texturepacks_path == self._path:
            raise ValueError("Could not find modules folder in '{}'.".format(self._path))
        return texturepacks_path.replace("\\", "/")

    def rims_path(
            self
    ) -> str:
        """
        Returns the path to rims folder of the Installation. This method maintains the case of the foldername.

        Returns:
            The path to the rims folder.
        """
        path = self._path
        for folder in os.listdir(self._path):
            if os.path.isdir(path + folder) and folder.lower() == "rims":
                path += folder + "/"
        if path == self._path:
            raise ValueError("Could not find rims folder in '{}'.".format(self._path))
        return path.replace("\\", "/")

    def streammusic_path(
            self
    ) -> str:
        """
        Returns the path to streammusic folder of the Installation. This method maintains the case of the foldername.

        Returns:
            The path to the streammusic folder.
        """
        path = self._path
        for folder in os.listdir(self._path):
            if os.path.isdir(path + folder) and folder.lower() == "streammusic":
                path += folder + "/"
        if path == self._path:
            raise ValueError("Could not find StreamMusic folder in '{}'.".format(self._path))
        return path.replace("\\", "/")

    def streamsounds_path(
            self
    ) -> str:
        """
        Returns the path to streamsounds folder of the Installation. This method maintains the case of the foldername.

        Returns:
            The path to the streamsounds folder.
        """
        path = self._path
        for folder in os.listdir(self._path):
            if os.path.isdir(path + folder) and folder.lower() == "streamsounds":
                path += folder + "/"
        if path == self._path:
            raise ValueError("Could not find StreamSounds folder in '{}'.".format(self._path))
        return path.replace("\\", "/")

    def streamvoice_path(
            self
    ) -> str:
        """
        Returns the path to streamvoice folder of the Installation. This method maintains the case of the foldername.

        In the first game, this folder has been named "streamwaves".

        Returns:
            The path to the streamvoice folder.
        """
        path = self._path
        for folder in os.listdir(self._path):
            if os.path.isdir(path + folder) and (folder.lower() == "streamvoice" or folder.lower() == "streamwaves"):
                path += folder + "/"
        if path == self._path:
            raise ValueError("Could not find voice over folder in '{}'.".format(self._path))
        return path.replace("\\", "/")
    # endregion

    # region Load Data
    def load_chitin(
            self
    ) -> None:
        """
        Reloads the list of resouces in the Chitin linked to the Installation.
        """
        chitin = Chitin(self._path)
        self._chitin = [resource for resource in chitin]

    def load_modules(
            self
    ) -> None:
        """
        Reloads the list of modules files in the modules folder linked to the Installation.
        """
        modules_path = self.module_path()
        self._modules = {}
        module_files = [file for file in os.listdir(modules_path) if file.endswith('.mod') or file.endswith('.rim') or file.endswith('.erf')]
        for module in module_files:
            with suppress(Exception):
                self._modules[module] = [resource for resource in Capsule(self.module_path() + module)]

    def reload_module(
            self,
            module: str
    ) -> None:
        """
        Reloads the list of resouces in specified module in the modules folder linked to the Installation.

        Args:
            module: The filename of the module.
        """
        self._modules[module] = [resource for resource in Capsule(self.module_path() + module)]

    def load_lips(
            self
    ) -> None:
        """
        Reloads the list of modules in the lips folder linked to the Installation.
        """
        self._lips = {}
        lips_path = self.lips_path()
        lip_files = [file for file in os.listdir(lips_path) if file.endswith('.mod')]
        for module in lip_files:
            self._lips[module] = [resource for resource in Capsule(lips_path + module)]

    def load_textures(
            self
    ) -> None:
        """
        Reloads the list of modules files in the texturepacks folder linked to the Installation.
        """
        self._texturepacks = {}
        texturepacks_path = self.texturepacks_path()
        texturepacks_files = [file for file in os.listdir(texturepacks_path) if file.endswith('.erf')]
        for module in texturepacks_files:
            self._texturepacks[module] = [resource for resource in Capsule(texturepacks_path + module)]

    def load_override(
            self
    ) -> None:
        """
        Reloads the list of subdirectories in the override folder linked to the Installation.
        """
        self._override = {}
        override_path = self.override_path()

        for path, subdirs, files in os.walk(override_path):
            directory = path.replace("\\", "/").replace(override_path, "")
            path = (path if path.endswith("/") else path + "/").replace("\\", "/")
            self._override[directory] = []

            for file in files:
                with suppress(Exception):
                    name, ext = file.split('.', 1)
                    size = os.path.getsize(path + file)
                    resource = FileResource(name, ResourceType.from_extension(ext), size, 0, path + file)
                    self._override[directory].append(resource)

    def reload_override(
            self,
            directory: str
    ) -> None:
        """
        Reloads the list of resources in subdirectory of the override folder linked to the Installation.

        Args:
            directory: The subdirectory in the override folder.
        """
        override_path = self.override_path()

        self._override[directory] = []
        files = os.listdir(override_path + directory)

        for file in files:
            with suppress(Exception):
                name, ext = file.split('.', 1)
                size = os.path.getsize(override_path + directory + file)
                resource = FileResource(name, ResourceType.from_extension(ext), size, 0, override_path + directory + file)
                self._override[directory].append(resource)

    def load_streammusic(
            self
    ) -> None:
        self._streammusic = []
        streammusic_path = self.streammusic_path()
        for filename in [file for file in os.listdir(streammusic_path)]:
            with suppress(Exception):
                filepath = streammusic_path + filename
                identifier = ResourceIdentifier.from_path(filepath)
                resource = FileResource(identifier.resname, identifier.restype, os.path.getsize(filepath), 0, filepath)
                self._streammusic.append(resource)

    def load_streamsounds(
            self
    ) -> None:
        """
        Reloads the list of resources in the streamsounds folder linked to the Installation.
        """
        self._streamsounds = []
        streamsounds_path = self.streamsounds_path()
        for filename in [file for file in os.listdir(streamsounds_path)]:
            with suppress(Exception):
                filepath = streamsounds_path + filename
                identifier = ResourceIdentifier.from_path(filepath)
                resource = FileResource(identifier.resname, identifier.restype, os.path.getsize(filepath), 0, filepath)
                self._streamsounds.append(resource)

    def load_streamvoices(
            self
    ) -> None:
        """
        Reloads the list of resources in the streamvoices folder linked to the Installation.
        """
        self._streamvoices = []
        streamvoices_path = self.streamvoice_path()

        for path, subdirs, files in os.walk(streamvoices_path):
            for filename in files:
                with suppress(Exception):
                    folderpath = path.replace("\\", "/")
                    if not folderpath.endswith("/"):
                        folderpath += "/"
                    filepath = folderpath + filename
                    identifier = ResourceIdentifier.from_path(filepath)
                    resource = FileResource(identifier.resname, identifier.restype, os.path.getsize(filepath), 0, filepath)
                    self._streamvoices.append(resource)

    def load_rims(
            self
    ) -> None:
        """
        Reloads the list of module files in the rims folder linked to the Installation.
        """
        self._rims = {}
        with suppress(ValueError):
            rims_path = self.rims_path()
            filenames = [file for file in os.listdir(rims_path) if file.endswith('.rim')]
            for filename in filenames:
                self._rims[filename] = [resource for resource in Capsule(rims_path + filename)]
    # endregion

    # region Get FileResources
    def chitin_resources(
            self
    ) -> List[FileResource]:
        """
        Returns the list of FileResources stored in the Chitin linked to the Installaiton.

        Returns:
            A list of FileResources.
        """
        return self._chitin[:]

    def modules_list(
            self
    ) -> List[str]:
        """
        Returns the list of module filenames located in the modules folder linked to the Installaiton.

        Module filenames are cached and require to be refreshed after a file is added, deleted or renamed.

        Returns:
            A list of filenames.
        """
        return list(self._modules.keys())

    def module_resources(
            self,
            filename: str
    ) -> List[FileResource]:
        """
        Returns a list of FileResources stored in the specified module file located in the modules folder linked to the
        Installation.

        Module resources are cached and require a reload after the contents have been modified.

        Returns:
            A list of FileResources.
        """
        return self._modules[filename][:]

    def lips_list(
            self
    ) -> List[str]:
        """
        Returns the list of module filenames located in the lips folder linked to the Installaiton.

        Module filenames are cached and require to be refreshed after a file is added, deleted or renamed.

        Returns:
            A list of filenames.
        """
        return list(self._lips.keys())

    def lip_resources(
            self,
            filename: str
    ) -> List[FileResource]:
        """
        Returns a list of FileResources stored in the specified module file located in the lips folder linked to the
        Installation.

        Module resources are cached and require a reload after the contents have been modified.

        Returns:
            A list of FileResources.
        """
        return self._lips[filename][:]

    def texturepacks_list(
            self
    ) -> List[str]:
        """
        Returns the list of texturepack filenames located in the texturepacks folder linked to the Installaiton.

        Returns:
            A list of filenames.
        """
        return list(self._texturepacks.keys())

    def texturepack_resources(
            self,
            filename: str
    ) -> List[FileResource]:
        """
        Returns a list of FileResources stored in the specified module file located in the texturepacks folder linked to
        the Installation.

        Texturepacks resources are cached and require a reload after the contents have been modified.

        Returns:
            A list of FileResources.
        """
        return self._texturepacks[filename][:]

    def override_list(
            self
    ) -> List[str]:
        """
        Returns the list of subdirectories located in override folder linked to the Installation.
        
        Subdirectories are cached and require to be refreshed after a folder is added, deleted or renamed.

        Returns:
            A list of subdirectories.
        """
        return list(self._override.keys())

    def override_resources(
            self,
            directory: str
    ) -> List[FileResource]:
        """
        Returns a list of FileResources stored in the specified subdirectory located in the override folder linked to
        the Installation.

        Override resources are cached and require a reload after the contents have been modified.

        Returns:
            A list of FileResources.
        """
        return list(self._override[directory])
    # endregion

    def talktable(
            self
    ) -> TalkTable:
        """
        Returns the TalkTable linked to the Installation.

        Returns:
            A TalkTable object.
        """
        return self._talktable

    def resource(
        self,
        resname: str,
        restype: ResourceType,
        order: List[SearchLocation] = None,
        *,
        capsules: List[Capsule] = None,
        folders: List[str] = None
    ) -> Optional[ResourceResult]:
        """
        Returns a resource matching the specified resref and restype. If no resource is found then None is returned
        instead.

        The default search order is (descending priority): 1. Folders in the folders parameter, 2. Override folders,
        3. Capsules in the capsules parameter, 4. Game modules, 5. Chitin.
        
        This is a wrapper of the resources() method provided to make fetching for a single resource more convienent.

        Args:
            resname: The name of the resource to look for.
            restype: The type of resource to look for.
            capsules: An extra list of capsules to search in.
            folders: An extra list of folders to search in.
            order: The ordered list of locations to check.

        Returns:
            A ResourceResult object if the specified resource is found, otherwise None.
        """

        query = ResourceIdentifier(resname, restype)
        batch = self.resources([query], order, capsules=capsules, folders=folders)

        return batch[query] if batch[query] else None

    def resources(
            self,
            queries: List[ResourceIdentifier],
            order: List[SearchLocation] = None,
            *,
            capsules: List[Capsule] = None,
            folders: List[str] = None
    ) -> Dict[ResourceIdentifier, Optional[ResourceResult]]:
        """
        Returns a dictionary mapping the items provided in the queries argument to the resource data if it was found. If
        the resource was not found, the value will be None.

        Args:
            queries: A list of resources to try load.
            order: The ordered list of locations to check.
            capsules: An extra list of capsules to search in.
            folders: An extra list of folders to search in.

        Returns:
            A dictionary mapping the given items in the queries argument to a list of ResourceResult objects.
        """
        results: Dict[ResourceIdentifier, Optional[ResourceResult]] = {}
        locations = self.locations(queries, order, capsules=capsules, folders=folders)
        handles = {}

        for query in queries:
            location = locations[query][0] if locations[query] else None
            if location is None:
                results[query] = None
            else:
                if query not in handles:
                    handles[query] = BinaryReader.from_file(location.filepath)
                handles[query].seek(location.offset)
                data = handles[query].read_bytes(location.size)
                results[query] = ResourceResult(query.resname, query.restype, location.filepath, data)

        for handle in handles.values():
            handle.close()

        return results

    def location(
            self,
            resname: str,
            restype: ResourceType,
            order: List[SearchLocation] = None,
            *,
            capsules: List[Capsule] = None,
            folders: List[str] = None
    ) -> List[LocationResult]:
        """
        Returns a list filepaths for where a particular resource matching the given resref and restype are located.

        This is a wrapper of the locations() method provided to make searching for a single resource more convienent.

        Args:
            resname: The name of the resource to look for.
            restype: The type of resource to look for.
            order: The ordered list of locations to check.
            capsules: An extra list of capsules to search in.
            folders: An extra list of folders to search in.

        Returns:
            A list of LocationResult objects.
        """
        capsules = [] if capsules is None else capsules
        folders = [] if folders is None else folders

        query: ResourceIdentifier = ResourceIdentifier(resname, restype)

        locations = self.locations([query], order, capsules=capsules, folders=folders)[query]

        return locations

    def locations(
            self,
            queries: List[ResourceIdentifier],
            order: List[SearchLocation] = None,
            *,
            capsules: List[Capsule] = None,
            folders: List[str] = None
    ) -> Dict[ResourceIdentifier, List[LocationResult]]:
        """
        Returns a dictionary mapping the items provided in the queries argument to a list of locations for that
        respective resource.

        Args:
            queries: A list of resources to try locate.
            order: The ordered list of locations to check.
            capsules: An extra list of capsules to search in.
            folders: An extra list of folders to search in.

        Returns:
            A dictionary mapping a resource identifier to a list of locations.
        """
        capsules = [] if capsules is None else capsules
        folders = [] if folders is None else folders

        order = [SearchLocation.CUSTOM_FOLDERS, SearchLocation.OVERRIDE, SearchLocation.CUSTOM_MODULES,
                 SearchLocation.MODULES, SearchLocation.CHITIN] if order is None else order

        locations: Dict[ResourceIdentifier, List[LocationResult]] = {}
        for qinden in queries:
            locations[qinden] = []

        def check_dict(values):
            for resources in values.values():
                for resource in resources:
                    if resource in queries:
                        location = LocationResult(resource.filepath(), resource.offset(), resource.size())
                        locations[resource.identifier()].append(location)

        def check_list(values):
            for resource in values:
                if resource in queries:
                    location = LocationResult(resource.filepath(), resource.offset(), resource.size())
                    locations[resource.identifier()].append(location)

        def check_capsules(values):
            for capsule in values:
                for query in queries:
                    if capsule.exists(query.resname, query.restype):
                        resource = capsule.info(query.resname, query.restype, reload=True)
                        location = LocationResult(resource.filepath(), resource.offset(), resource.size())
                        locations[resource.identifier()].append(location)

        def check_folders(values):
            for folder in values:
                folder = folder + '/' if not folder.endswith('/') else folder
                for file in [file for file in os.listdir(folder) if os.path.isfile(folder + file)]:
                    filepath = folder + file
                    for query in queries:
                        with suppress(Exception):
                                identifier = ResourceIdentifier.from_path(file)
                                if query == identifier:
                                    resource = FileResource(query.resname, query.restype, os.path.getsize(filepath), 0, filepath)
                                    location = LocationResult(resource.filepath(), resource.offset(), resource.size())
                                    locations[identifier].append(location)

        function_map = {
            SearchLocation.OVERRIDE: lambda: check_dict(self._override),
            SearchLocation.MODULES: lambda: check_dict(self._modules),
            SearchLocation.LIPS: lambda: check_dict(self._lips),
            SearchLocation.RIMS: lambda: check_dict(self._rims),
            SearchLocation.TEXTURES_TPA: lambda: check_list(self._texturepacks["swpc_tex_tpa.erf"]),
            SearchLocation.TEXTURES_TPB: lambda: check_list(self._texturepacks["swpc_tex_tpb.erf"]),
            SearchLocation.TEXTURES_TPC: lambda: check_list(self._texturepacks["swpc_tex_tpc.erf"]),
            SearchLocation.TEXTURES_GUI: lambda: check_list(self._texturepacks["swpc_tex_gui.erf"]),
            SearchLocation.CHITIN: lambda: check_list(self._chitin),
            SearchLocation.MUSIC: lambda: check_list(self._streammusic),
            SearchLocation.SOUND: lambda: check_list(self._streamsounds),
            SearchLocation.VOICE: lambda: check_list(self._streamvoices),
            SearchLocation.CUSTOM_MODULES: lambda: check_capsules(capsules),
            SearchLocation.CUSTOM_FOLDERS: lambda: check_folders(folders)
        }

        for item in order:
            function_map[item]()

        return locations

    def texture(
            self,resname: str,
            order: List[SearchLocation] = None,
            *,
            capsules: List[Capsule] = None,
            folders: List[str] = None
    ) -> Optional[TPC]:
        """
        Returns a TPC object loaded from a resource with the specified name. If the specified texture could not be found
        then the method returns None.

        This is a wrapper of the textures() method provided to make searching for a single texture more convienent.

        Texture is search for in the following order:
            1. "folders" parameter.
            2. "capsules" parameter.
            3. Installation override folder.
            4. Normal texture pack.
            5. GUI texture pack.
            6. Installation Chitin.
            7. Installation module files in modules folder.

        Args:
            resname: The ResRef string.
            order: The ordered list of locations to check.
            capsules: An extra list of capsules to search in.
            folders: An extra list of folders to search in.

        Returns:
            TPC object or None.
        """

        batch = self.textures([resname], order, capsules=capsules, folders=folders)
        return batch[resname] if batch else None

    def textures(
            self,
            queries: List[str],
            order: List[SearchLocation] = None,
            *,
            capsules: List[Capsule] = None,
            folders: List[str] = None
    ) -> CaseInsensitiveDict[Optional[TPC]]:
        """
        Returns a dictionary mapping the items provided in the queries argument to a TPC object if it exists. If the
        texture could not be found then the value is None.

        Args:
            queries: A list of resources to try locate.
            order: The ordered list of locations to check.
            capsules: An extra list of capsules to search in.
            folders: An extra list of folders to search in.

        Returns:
            A dictionary mapping case-insensitive strings to TPC objects or None.
        """
        capsules = [] if capsules is None else capsules
        folders = [] if folders is None else folders

        order = [SearchLocation.CUSTOM_FOLDERS, SearchLocation.OVERRIDE, SearchLocation.CUSTOM_MODULES,
                 SearchLocation.TEXTURES_TPA, SearchLocation.CHITIN] if order is None else order

        textures: CaseInsensitiveDict[Optional[TPC]] = CaseInsensitiveDict[Optional[TPC]]()
        texture_types = [ResourceType.TPC, ResourceType.TGA]
        queries = [resname.lower() for resname in queries]

        for resname in queries:
            textures[resname] = None

        def check_dict(values):
            for resources in values.values():
                for resource in resources:
                    if resource.resname() in copy(queries) and resource.restype() in texture_types:
                        queries.remove(resource.resname())
                        textures[resource.resname()] = read_tpc(resource.data())

        def check_list(values):
            for resource in values:
                if resource.resname() in copy(queries) and resource.restype() in texture_types:
                    queries.remove(resource.resname())
                    textures[resource.resname()] = read_tpc(resource.data())

        def check_capsules(values):
            for capsule in values:
                for resname in queries:
                    if capsule.exists(resname, ResourceType.TPC):
                        queries.remove(resname)
                        textures[resname] = read_tpc(capsule.resource(resname, ResourceType.TPC))
                    if capsule.exists(resname, ResourceType.TGA):
                        queries.remove(resname)
                        textures[resname] = read_tpc(capsule.resource(resname, ResourceType.TGA))

        def check_folders(values):
            for folder in values:
                folder = folder + '/' if not folder.endswith('/') else folder
                for file in [file for file in os.listdir(folder) if os.path.isfile(folder + file)]:
                    filepath = folder + file
                    identifier = ResourceIdentifier.from_path(file)
                    for resname in queries:
                        if identifier.resname == resname and identifier.restype in texture_types:
                            data = BinaryReader.load_file(filepath)
                            textures[resname] = read_tpc(data)

        function_map = {
            SearchLocation.OVERRIDE: lambda: check_dict(self._override),
            SearchLocation.MODULES: lambda: check_dict(self._modules),
            SearchLocation.LIPS: lambda: check_dict(self._lips),
            SearchLocation.RIMS: lambda: check_dict(self._rims),
            SearchLocation.TEXTURES_TPA: lambda: check_list(self._texturepacks["swpc_tex_tpa.erf"]),
            SearchLocation.TEXTURES_TPB: lambda: check_list(self._texturepacks["swpc_tex_tpb.erf"]),
            SearchLocation.TEXTURES_TPC: lambda: check_list(self._texturepacks["swpc_tex_tpc.erf"]),
            SearchLocation.TEXTURES_GUI: lambda: check_list(self._texturepacks["swpc_tex_gui.erf"]),
            SearchLocation.CHITIN: lambda: check_list(self._chitin),
            SearchLocation.MUSIC: lambda: check_list(self._streammusic),
            SearchLocation.SOUND: lambda: check_list(self._streamsounds),
            SearchLocation.VOICE: lambda: check_list(self._streamvoices),
            SearchLocation.CUSTOM_MODULES: lambda: check_capsules(capsules),
            SearchLocation.CUSTOM_FOLDERS: lambda: check_folders(folders)
        }

        for item in order:
            function_map[item]()

        return textures

    def sound(
            self,
            queries: str,
            order: List[SearchLocation] = None,
            *,
            capsules: List[Capsule] = None,
            folders: List[str] = None
    ) -> Optional[Optional[bytes]]:
        """
        Returns the bytes of a sound resource if it can be found, otherwise returns None.

        This is a wrapper of the sounds() method provided to make searching for a single resource more convienent.

        Args:
            queries: The name of the resource to look for.
            order: The ordered list of locations to check.
            capsules: An extra list of capsules to search in.
            folders: An extra list of folders to search in.

        Returns:
            A bytes object or None.
        """
        batch = self.sounds([queries], order, capsules=capsules, folders=folders)
        return batch[queries] if batch else None

    def sounds(
            self,
            resnames: List[str],
            order: List[SearchLocation] = None,
            *,
            capsules: List[Capsule] = None,
            folders: List[str] = None
    ) -> CaseInsensitiveDict[Optional[bytes]]:
        """
        Returns a dictionary mapping the items provided in the queries argument to a bytes object if the respective
        sound resource could be found. If the sound could not be found the value will return None.

        Args:
            resnames: A list of sounds to try locate.
            order: The ordered list of locations to check.
            capsules: An extra list of capsules to search in.
            folders: An extra list of folders to search in.

        Returns:
            A dictionary mapping a case-insensitive string to a bytes object or None.
        """
        capsules = [] if capsules is None else capsules
        folders = [] if folders is None else folders

        order = [SearchLocation.CUSTOM_FOLDERS, SearchLocation.OVERRIDE, SearchLocation.CUSTOM_MODULES,
                 SearchLocation.SOUND, SearchLocation.CHITIN] if not order else order

        sounds: CaseInsensitiveDict[Optional[bytes]] = CaseInsensitiveDict[Optional[bytes]]()
        texture_types = [ResourceType.WAV, ResourceType.MP3]
        resnames = [resname.lower() for resname in resnames]

        for resname in resnames:
            sounds[resname] = None

        def check_dict(values):
            for resources in values.values():
                for resource in resources:
                    if resource.resname() in copy(resnames) and resource.restype() in texture_types:
                        resnames.remove(resource.resname())
                        sounds[resource.resname()] = sound.fix_audio(resource.data())

        def check_list(values):
            for resource in values:
                if resource.resname() in copy(resnames) and resource.restype() in texture_types:
                    resnames.remove(resource.resname())
                    sounds[resource.resname()] = sound.fix_audio(resource.data())

        def check_capsules(values):
            for capsule in values:
                for resname in resnames:
                    if capsule.exists(resname, ResourceType.WAV):
                        resnames.remove(resname)
                        sounds[resname] = sound.fix_audio(capsule.resource(resname, ResourceType.TPC))
                    if capsule.exists(resname, ResourceType.MP3):
                        resnames.remove(resname)
                        sounds[resname] = sound.fix_audio(capsule.resource(resname, ResourceType.TGA))

        def check_folders(values):
            for folder in values:
                folder = folder + '/' if not folder.endswith('/') else folder
                for file in [file for file in os.listdir(folder) if os.path.isfile(folder + file)]:
                    filepath = folder + file
                    identifier = ResourceIdentifier.from_path(file)
                    for resname in resnames:
                        if identifier.resname == resname and identifier.restype in texture_types:
                            data = BinaryReader.load_file(filepath)
                            sounds[resname] = sound.fix_audio(data)

        function_map = {
            SearchLocation.OVERRIDE: lambda: check_dict(self._override),
            SearchLocation.MODULES: lambda: check_dict(self._modules),
            SearchLocation.LIPS: lambda: check_dict(self._lips),
            SearchLocation.RIMS: lambda: check_dict(self._rims),
            SearchLocation.TEXTURES_TPA: lambda: check_list(self._texturepacks["swpc_tex_tpa.erf"]),
            SearchLocation.TEXTURES_TPB: lambda: check_list(self._texturepacks["swpc_tex_tpb.erf"]),
            SearchLocation.TEXTURES_TPC: lambda: check_list(self._texturepacks["swpc_tex_tpc.erf"]),
            SearchLocation.TEXTURES_GUI: lambda: check_list(self._texturepacks["swpc_tex_gui.erf"]),
            SearchLocation.CHITIN: lambda: check_list(self._chitin),
            SearchLocation.MUSIC: lambda: check_list(self._streammusic),
            SearchLocation.SOUND: lambda: check_list(self._streamsounds),
            SearchLocation.VOICE: lambda: check_list(self._streamvoices),
            SearchLocation.CUSTOM_MODULES: lambda: check_capsules(capsules),
            SearchLocation.CUSTOM_FOLDERS: lambda: check_folders(folders)
        }

        for item in order:
            function_map[item]()

        return sounds

    def string(
            self,
            locstring: LocalizedString,
            default: str = ""
    ) -> str:
        """
        Returns the string for the LocalizedString provided.
        
        This is a wrapper of the strings() method provided to make searching for a single string more convienent.
        
        Args:
            locstring: 
            default: 

        Returns:

        """
        batch = self.strings([locstring], default)
        return batch[locstring]

    def strings(
            self,
            queries: List[LocalizedString],
            default: str = ""
    ) -> Dict[LocalizedString, str]:
        """
        Returns a dictionary mapping the items provided in the queries argument to a string.
        
        As the method iterates through each LocalizedString it will first check if the TalkTable linked to the
        Installation has the stringref. If not it will try fallback on whatever substring exists in the LocalizedString
        and should that fail it will fallback on the default string specified.
        
        Args:
            queries: A list of LocalizedStrings.
            default: The fallback string if no string could be found.

        Returns:
            A dictionary mapping LocalizedString to a string.
        """
        stringrefs = [locstring.stringref for locstring in queries]
        batch = self.talktable().batch(stringrefs)

        results = {}
        for locstring in queries:
            if locstring.stringref in batch and locstring.stringref != -1:
                results[locstring] = batch[locstring.stringref].text
            elif len(locstring):
                for language, gender, text in locstring:
                    results[locstring] = text
                    break
            else:
                results[locstring] = default

        return results

    def module_name(
            self,
            module_filename: str,
            use_hardcoded: bool = True
    ) -> str:
        """
        Returns the name of the area for a module from the installations module list. The name is taken from the
        LocalizedString "Name" in the relevant module file's ARE resource.

        Args:
            module_filename: The name of the module file.
            use_hardcoded: Use hardcoded values for modules where applicable.

        Returns:
            The name of the area for the module.
        """
        root = module_filename.replace(".mod", "").replace(".erf", "").replace(".rim", "")
        root = root[:-len("_s")] if root.endswith("_s") else root
        root = root[:-len("_dlg")] if root.endswith("_dlg") else root

        hardcoded = {
            "STUNT_00": "Ebon Hawk - Cutscene (Vision Sequences)",
            "STUNT_03A": "Leviathan - Cutscene (Destroy Taris)",
            "STUNT_06": "Leviathan - Cutscene (Resume Bombardment)",
            "STUNT_07": "Ebon Hawk - Cutscene (Escape Taris)",
            "STUNT_12": "Leviathan - Cutscene (Calo Nord)",
            "STUNT_14": "Leviathan - Cutscene (Darth Bandon)",
            "STUNT_16": "Ebon Hawk - Cutscene (Leviathan Capture)",
            "STUNT_18": "Unknown World - Cutscene (Bastila Torture)",
            "STUNT_19": "Star Forge - Cutscene (Jawless Malak)",
            "STUNT_31B": "Unknown World - Cutscene (Revan Reveal)",
            "STUNT_34": "Ebon Hawk - Cutscene (Star Forge Arrival)",
            "STUNT_35": "Ebon Hawk - Cutscene (Lehon Crash)",
            "STUNT_42": "Ebon Hawk - Cutscene (LS Dodonna Call)",
            "STUNT_44": "Ebon Hawk - Cutscene (DS Dodonna Call)",
            "STUNT_50A": "Dodonna Flagship - Cutscene (Break In Formation)",
            "STUNT_51A": "Dodonna Flagship - Cutscene (Bastilla Against Us)",
            "STUNT_54A": "Dodonna Flagship - Cutscene (Pull Back)",
            "STUNT_55A": "Unknown World - Cutscene (DS Ending)",
            "STUNT_56A": "Dodona Flagship - Cutscene (Star Forge Destroyed)",
            "STUNT_57": "Unknown World - Cutscene (LS Ending)",
            "001EBO": "Ebon Hawk - Interior (Prologue)",
            "004EBO": "Ebon Hawk - Interior (Red Eclipse)",
            "005EBO": "Ebon Hawk - Interior (Escaping Peragus)",
            "006EBO": "Ebon Hawk - Cutscene (After Rebuilt Enclave)",
            "007EBO": "Ebon Hawk - Cutscene (After Goto's Yatch)",
            "154HAR": "Harbinger - Cutscene (Sion Introduction)",
            "205TEL": "Citadel Station - Cutscene (Carth Discussion)",
            "352NAR": "Nar Shaddaa - Cutscene (Goto Introduction)",
            "853NIH": "Ravager - Cutscene (Nihilus Introduction)",
            "856NIH": "Ravager - Cutscene (Sion vs. Nihilus)"
        }

        if use_hardcoded:
            for key in hardcoded.keys():
                if key.upper() in module_filename.upper():
                    return hardcoded[key]

        name = ""

        for module in self.modules_list():
            if root not in module:
                continue

            capsule = Capsule(self.module_path() + module)
            tag = ""

            if capsule.exists("module", ResourceType.IFO):
                ifo = read_gff(capsule.resource("module", ResourceType.IFO))
                tag = ifo.root.get_resref("Mod_Entry_Area").get()
            if capsule.exists(tag, ResourceType.ARE):
                are = read_gff(capsule.resource(tag, ResourceType.ARE))
                locstring = are.root.get_locstring("Name")
                if locstring.stringref > 0:
                    name = self._talktable.string(locstring.stringref)
                elif locstring.exists(Language.ENGLISH, Gender.MALE):
                    name = locstring.get(Language.ENGLISH, Gender.MALE)
                break

        return name

    def module_names(
            self
    ) -> Dict[str, str]:
        """
        Returns a dictionary mapping module filename to the name of the area. The name is taken from the LocalizedString
        "Name" in the relevant module file's ARE resource.

        Returns:
            A dictionary mapping module filename to in-game module area name.
        """
        module_names = {}
        for module in self.modules_list():
            module_names[module] = self.module_name(module)
        return module_names

    def module_id(
            self,
            module_filename: str,
            use_hardcoded: bool = True
    ) -> str:
        """
        Returns the ID of the area for a module from the installations module list. The ID is taken from the
        ResRef field "Mod_Entry_Area" in the relevant module file's IFO resource.

        Args:
            module_filename: The name of the module file.
            use_hardcoded: Use hardcoded values for modules where applicable.

        Returns:
            The ID of the area for the module.
        """
        root = module_filename.replace(".mod", "").replace(".erf", "").replace(".rim", "")
        root = root[:-len("_s")] if root.endswith("_s") else root
        root = root[:-len("_dlg")] if root.endswith("_dlg") else root

        hardcoded = {
            "STUNT_00": "000",
            "STUNT_03A": "m03a",
            "STUNT_06": "m07",
            "STUNT_07": "m07",
            "STUNT_12": "m12",
            "STUNT_14": "m14",
            "STUNT_16": "m16",
            "STUNT_18": "m18",
            "STUNT_19": "m19",
            "STUNT_31B": "m31b",
            "STUNT_34": "m34",
            "STUNT_35": "m35",
            "STUNT_42": "m43",
            "STUNT_44": "m44",
            "STUNT_50A": "m50a",
            "STUNT_51A": "m51a",
            "STUNT_54A": "m54a",
            "STUNT_55A": "m55a",
            "STUNT_56A": "m56a",
            "STUNT_57": "m57",
        }

        if use_hardcoded:
            for key in hardcoded.keys():
                if key.upper() in module_filename.upper():
                    return hardcoded[key]

        mod_id = ""

        for module in self.modules_list():
            if root not in module:
                continue

            capsule = Capsule(self.module_path() + module)

            if capsule.exists("module", ResourceType.IFO):
                ifo = read_gff(capsule.resource("module", ResourceType.IFO))
                mod_id = ifo.root.get_resref("Mod_Entry_Area").get()

        return mod_id

    def module_ids(
            self
    ) -> Dict[str, str]:
        """
        Returns a dictionary mapping module filename to the ID of the module. The ID is taken from the
        ResRef field "Mod_Entry_Area" in the relevant module file's IFO resource.

        Returns:
            A dictionary mapping module filename to in-game module id.
        """
        module_ids = {}
        for module in self.modules_list():
            module_ids[module] = self.module_id(module)
        return module_ids

    def uninstall_mods(
            self
    ) -> None:
        """
        Uninstalls all mods from the game.

        What this method really does is delete all the contents of the override folder and delete all .MOD files from
        the modules folder.
        """
        for file in os.listdir(self.module_path()):
            if file.endswith(".mod") and os.path.isfile(self.module_path() + file):
                os.remove(self.module_path() + file)

        for file in os.listdir(self.override_path()):
            if os.path.isfile(self.override_path() + file):
                os.remove(self.override_path() + file)



"""
This module handles classes relating to editing TPC files.
"""
from __future__ import annotations

from enum import IntEnum
from typing import List, Tuple, Optional, NamedTuple

from pykotor.common.stream import BinaryReader
from pykotor.resource.type import ResourceType


class TPCGetResult(NamedTuple):
    width: int
    height: int
    format: TPCTextureFormat
    data: Optional[bytes]


class TPCConvertResult(NamedTuple):
    width: int
    height: int
    data: Optional[bytes]


class TPC:
    """
    Represents a TPC file.

    Attributes:
        txi: Stores additional information regarding the texture.
    """

    BINARY_TYPE = ResourceType.TPC

    def __init__(
            self
    ):
        self._texture_format: TPCTextureFormat = TPCTextureFormat.RGB
        self._mipmaps: List[bytes] = [bytes([0 for i in range(4 * 4 * 3)])]
        self._width: int = 4
        self._height: int = 4
        self.txi: str = ""

        # TODO: cube maps

    def mipmap_count(
            self
    ) -> int:
        """
        Returns the number of mipmaps.

        Returns:
            The number of mipmaps.
        """
        return len(self._mipmaps)

    def format(
            self
    ) -> TPCTextureFormat:
        """
        Returns the format of the stored texture.

        Returns:
            The format of the stored texture.
        """
        return self._texture_format

    def dimensions(
            self
    ) -> Tuple[int, int]:
        """
        Returns the width and height of the largest mipmap.

        Returns:
            A tuple containing [width, height].
        """
        return self._width, self._height

    def get(
            self,
            mipmap: int = 0
    ) -> TPCGetResult:
        """
        Returns a tuple containing the width, height, texture format, and data of the specified mipmap.

        Args:
            mipmap: The index of the mipmap.

        Returns:
            A tuple equal to (width, height, texture format, data).
        """
        width, height = self._mipmap_size(mipmap)
        return TPCGetResult(width, height, self._texture_format, self._mipmaps[mipmap])

    def convert(
            self,
            convert_format: TPCTextureFormat,
            mipmap: int = 0
    ) -> TPCConvertResult:
        """
        Returns a tuple containing the width, height and data of the specified mipmap where the data returned is in
        the texture format specified.

        Args:
            convert_format: The format the texture data should be converted to.
            mipmap: The index of the mipmap.

        Returns:
            A tuple equal to (width, height, data)
        """
        width, height = self._mipmap_size(mipmap)
        raw_data = self._mipmaps[mipmap]
        data = bytes()

        if convert_format == TPCTextureFormat.DXT1 or convert_format == TPCTextureFormat.DXT5:
            raise NotImplementedError()

        if convert_format == TPCTextureFormat.Greyscale:
            raise NotImplementedError()

        if convert_format == TPCTextureFormat.RGBA:
            if self._texture_format == TPCTextureFormat.DXT5:
                data = TPC._dxt5_to_rgba(raw_data, width, height)
            elif self._texture_format == TPCTextureFormat.DXT1:
                data = TPC._dxt1_to_rgba(raw_data, width, height)
            elif self._texture_format == TPCTextureFormat.RGBA:
                data = raw_data
            elif self._texture_format == TPCTextureFormat.RGB:
                data = TPC._rgb_to_rgba(raw_data, width, height)
            elif self._texture_format == TPCTextureFormat.Greyscale:
                data = TPC._grey_to_rgba(raw_data, width, height)

        if convert_format == TPCTextureFormat.RGB:
            if self._texture_format == TPCTextureFormat.DXT5:
                data = TPC._dxt5_to_rgb(raw_data, width, height)
            elif self._texture_format == TPCTextureFormat.DXT1:
                data = TPC._dxt1_to_rgb(raw_data, width, height)
            elif self._texture_format == TPCTextureFormat.RGBA:
                data = TPC._rgba_to_rgb(raw_data, width, height)
            elif self._texture_format == TPCTextureFormat.RGB:
                data = raw_data
            elif self._texture_format == TPCTextureFormat.Greyscale:
                data = TPC._grey_to_rgba(raw_data, width, height)
                data = TPC._rgba_to_rgb(data, width, height)

        return TPCConvertResult(width, height, data)

    def set_single(
            self,
            width: int,
            height: int,
            data: bytes,
            texture_format: TPCTextureFormat
    ) -> None:
        """
        Sets the texture data but only for a single mipmap.

        Args:
            width: The new width.
            height: The new height.
            data: The new texture data.
            texture_format: The texture format.
        """
        self.set(width, height, [data], texture_format)

    def set(
            self,
            width: int,
            height: int,
            mipmaps: List[bytes],
            texture_format: TPCTextureFormat
    ) -> None:
        """
        Sets the new texture data.

        Args:
            width: The new width.
            height: The new height.
            mipmaps: The new mipmaps data.
            texture_format: The texture format.
        """
        # TODO: Some sort of simple sanity check on the data; make sure the mipmaps' data have the appropriate size
        #       according to their texture format.

        self._texture_format = texture_format
        self._mipmaps = mipmaps
        self._width = width
        self._height = height

    def is_compressed(
            self
    ) -> bool:
        return self._texture_format in [TPCTextureFormat.DXT1, TPCTextureFormat.DXT5]

    def _mipmap_size(
            self,
            mipmap: int
    ) -> Tuple[int, int]:
        """
        Returns the size of the specified mipmap.

        Args:
            mipmap: The index of the mipmap.

        Raises:
            IndexError: The index for the mipmap is out of range.

        Returns:
            A tuple equal to (width, height).
        """
        if 0 > mipmap >= len(self._mipmaps):
            raise IndexError("The index for the mipmap is out of range.")

        width = self._width
        height = self._height
        for i in range(mipmap):
            width >>= 1
            height >>= 1
        return width, height

    # region Convert to RGBA
    @staticmethod
    def _dxt5_to_rgba(
            data: bytes,
            width: int,
            height: int
    ) -> bytearray:
        dxt_reader = BinaryReader.from_bytes(data)
        new_data = bytearray(width * height * 4)

        for ty in range(4, height + 4, 4):
            for tx in range(0, width, 4):
                alpha0 = dxt_reader.read_uint8()
                alpha1 = dxt_reader.read_uint8()
                dxt_alpha = TPC._integer48(dxt_reader.read_bytes(6))

                x = dxt_reader.read_int16()
                y = dxt_reader.read_int16()
                c0 = TPC._rgba565_to_rgb(x)
                c1 = TPC._rgba565_to_rgb(y)
                dxt_pixels = dxt_reader.read_uint32(big=True)

                cc = [c0, c1]
                if TPC._rgba565_to_rgb888(x) > TPC._rgba565_to_rgb888(y):
                    cc.extend([TPC._interpolate_rgb(0.3333333, c0, c1), TPC._interpolate_rgb(0.6666666, c0, c1)])
                else:
                    cc.extend([TPC._interpolate_rgb(0.5555555, c0, c1), (0, 0, 0)])

                alpha_code = [alpha0, alpha1]
                if alpha0 > alpha1:
                    alpha_code.append(int((6.0 * alpha0 + 1.0 * alpha1 + 3) / 7))
                    alpha_code.append(int((5.0 * alpha0 + 2.0 * alpha1 + 3) / 7))
                    alpha_code.append(int((4.0 * alpha0 + 3.0 * alpha1 + 3) / 7))
                    alpha_code.append(int((3.0 * alpha0 + 4.0 * alpha1 + 3) / 7))
                    alpha_code.append(int((2.0 * alpha0 + 5.0 * alpha1 + 3) / 7))
                    alpha_code.append(int((1.0 * alpha0 + 6.0 * alpha1 + 3) / 7))
                else:
                    alpha_code.append(int((4.0 * alpha0 + 1.0 * alpha1 + 1) / 5))
                    alpha_code.append(int((3.0 * alpha0 + 2.0 * alpha1 + 2) / 5))
                    alpha_code.append(int((2.0 * alpha0 + 3.0 * alpha1 + 2) / 5))
                    alpha_code.append(int((1.0 * alpha0 + 4.0 * alpha1 + 2) / 5))
                    alpha_code.append(0)
                    alpha_code.append(255)

                for y in [3, 2, 1, 0]:
                    for x in [0, 1, 2, 3]:
                        pixelc_code = dxt_pixels & 3
                        dxt_pixels >>= 2

                        a = alpha_code[(dxt_alpha >> (3 * (4 * y + x))) & 7]

                        index = ((ty - 4 + y) * width + (tx + x)) * 4
                        new_data[index + 0] = cc[pixelc_code][0]
                        new_data[index + 1] = cc[pixelc_code][1]
                        new_data[index + 2] = cc[pixelc_code][2]
                        new_data[index + 3] = a

        return new_data

    @staticmethod
    def _dxt1_to_rgba(
            data: bytes,
            width: int,
            height: int
    ) -> bytearray:
        dxt_reader = BinaryReader.from_bytes(data)
        new_data = bytearray(width * height * 4)

        for ty in range(4, height + 4, 4):
            for tx in range(0, width, 4):
                x = dxt_reader.read_int16()
                y = dxt_reader.read_int16()
                c0 = TPC._rgba565_to_rgb(x)
                c1 = TPC._rgba565_to_rgb(y)
                dxt_pixels = dxt_reader.read_uint32(big=True)

                cc = [c0, c1]
                if TPC._rgba565_to_rgb888(x) > TPC._rgba565_to_rgb888(y):
                    cc.extend([TPC._interpolate_rgb(0.3333333, c0, c1), TPC._interpolate_rgb(0.6666666, c0, c1)])
                else:
                    cc.extend([TPC._interpolate_rgb(0.5555555, c0, c1), (0, 0, 0)])

                for y in [3, 2, 1, 0]:
                    for x in [0, 1, 2, 3]:
                        pixelc_code = dxt_pixels & 3
                        dxt_pixels >>= 2

                        index = ((ty - 4 + y) * width + (tx + x)) * 4
                        new_data[index + 0] = cc[pixelc_code][0]
                        new_data[index + 1] = cc[pixelc_code][1]
                        new_data[index + 2] = cc[pixelc_code][2]
                        new_data[index + 3] = 255

        return new_data

    @staticmethod
    def _rgb_to_rgba(
            data: bytes,
            width: int,
            height: int
    ) -> bytearray:
        new_data = bytearray()
        rgb_reader = BinaryReader.from_bytes(data)

        for ty in range(4, height + 4, 4):
            for x in range(width):
                new_data.extend([rgb_reader.read_uint8(), rgb_reader.read_uint8(), rgb_reader.read_uint8(), 255])

        return new_data

    @staticmethod
    def _grey_to_rgba(
            data: bytes,
            width: int,
            height: int
    ) -> bytearray:
        new_data = bytearray()
        rgb_reader = BinaryReader.from_bytes(data)

        for y in range(height):
            for x in range(width):
                brightness = rgb_reader.read_uint8()
                new_data.extend([brightness, brightness, brightness, 255])

        return new_data

    # endregion

    # region Convert to Grey
    @staticmethod
    def _rgba_to_grey(
            data: bytes,
            width: int,
            height: int
    ) -> bytearray:
        new_data = bytearray()
        rgb_reader = BinaryReader.from_bytes(data)

        for y in range(height):
            for x in range(width):
                r = rgb_reader.read_uint8()
                g = rgb_reader.read_uint8()
                b = rgb_reader.read_uint8()
                a = rgb_reader.read_uint8()
                highest = r
                if g > highest:
                    highest = g
                if b > highest:
                    highest = b
                new_data.extend([highest])

        return new_data

    # endregion

    # region Convert to RGB
    @staticmethod
    def _dxt5_to_rgb(
            data: bytes,
            width: int,
            height: int
    ) -> bytearray:
        dxt_reader = BinaryReader.from_bytes(data)
        new_data = bytearray(width * height * 3)

        for ty in range(4, height + 4, 4):
            for tx in range(0, width, 4):
                dxt_reader.skip(8)

                x = dxt_reader.read_int16()
                y = dxt_reader.read_int16()
                c0 = TPC._rgba565_to_rgb(x)
                c1 = TPC._rgba565_to_rgb(y)
                dxt_pixels = dxt_reader.read_uint32(big=True)

                cc = [c0, c1]
                if TPC._rgba565_to_rgb888(x) > TPC._rgba565_to_rgb888(y):
                    cc.extend([TPC._interpolate_rgb(0.3333333, c0, c1), TPC._interpolate_rgb(0.6666666, c0, c1)])
                else:
                    cc.extend([TPC._interpolate_rgb(0.5555555, c0, c1), (0, 0, 0)])

                for y in [3, 2, 1, 0]:
                    for x in [0, 1, 2, 3]:
                        pixelc_code = dxt_pixels & 3
                        dxt_pixels >>= 2

                        index = ((ty - 4 + y) * width + (tx + x)) * 3
                        new_data[index + 0] = cc[pixelc_code][0]
                        new_data[index + 1] = cc[pixelc_code][1]
                        new_data[index + 2] = cc[pixelc_code][2]

        return new_data

    @staticmethod
    def _dxt1_to_rgb(
            data: bytes,
            width: int,
            height: int
    ) -> bytearray:
        dxt_reader = BinaryReader.from_bytes(data)
        new_data = bytearray(width * height * 3)

        for ty in range(4, height + 4, 4):
            for tx in range(0, width, 4):
                x = dxt_reader.read_int16()
                y = dxt_reader.read_int16()
                c0 = TPC._rgba565_to_rgb(x)
                c1 = TPC._rgba565_to_rgb(y)
                dxt_pixels = dxt_reader.read_uint32(big=True)

                cc = [c0, c1]
                if TPC._rgba565_to_rgb888(x) > TPC._rgba565_to_rgb888(y):
                    cc.extend([TPC._interpolate_rgb(0.3333333, c0, c1), TPC._interpolate_rgb(0.6666666, c0, c1)])
                else:
                    cc.extend([TPC._interpolate_rgb(0.5555555, c0, c1), (0, 0, 0)])

                for y in [3, 2, 1, 0]:
                    for x in [0, 1, 2, 3]:
                        pixelc_code = dxt_pixels & 3
                        dxt_pixels >>= 2

                        index = ((ty - 4 + y) * width + (tx + x)) * 3
                        new_data[index + 0] = cc[pixelc_code][0]
                        new_data[index + 1] = cc[pixelc_code][1]
                        new_data[index + 2] = cc[pixelc_code][2]

        return new_data

    @staticmethod
    def _rgba_to_rgb(
            data: bytes,
            width: int,
            height: int
    ) -> bytearray:
        new_data = bytearray()
        rgb_reader = BinaryReader.from_bytes(data)

        for y in range(height):
            for x in range(width):
                new_data.extend([rgb_reader.read_uint8(), rgb_reader.read_uint8(), rgb_reader.read_uint8()])
                rgb_reader.skip(1)

        return new_data

    # endregion

    @staticmethod
    def _rgba565_to_rgb888(
            color: int
    ) -> int:
        blue = (color & 0x1F)
        green = (color >> 5) & 0x3F
        red = (color >> 11) & 0x1F
        return (blue << 3) + (green << 10) + (red << 19)

    @staticmethod
    def _interpolate(
            weight: float,
            color0: int,
            color1: int
    ) -> int:
        color0_blue = color0 & 255
        color0_greed = (color0 >> 8) & 255
        color0_red = (color0 >> 16) & 255

        color1_blue = color1 & 255
        color1_greed = (color1 >> 8) & 255
        color1_red = (color1 >> 16) & 255

        blue = int(((1.0 - weight) * color0_blue) + (weight * color1_blue))
        green = int(((1.0 - weight) * color0_greed) + (weight * color1_greed))
        red = int(((1.0 - weight) * color0_red) + (weight * color1_red))

        return blue + (green << 8) + (red << 16)

    @staticmethod
    def _rgba565_to_rgb(
            color: int
    ):
        blue = color & 0x1F
        green = (color >> 5) & 0x3F
        red = (color >> 11) & 0x1F
        return red << 3, green << 2, blue << 3

    @staticmethod
    def _interpolate_rgb(
            weight: float,
            color0,
            color1
    ):
        color0_blue = color0[2]
        color0_greed = color0[1]
        color0_red = color0[0]

        color1_blue = color1[2]
        color1_greed = color1[1]
        color1_red = color1[0]

        blue = int(((1.0 - weight) * color0_blue) + (weight * color1_blue))
        green = int(((1.0 - weight) * color0_greed) + (weight * color1_greed))
        red = int(((1.0 - weight) * color0_red) + (weight * color1_red))

        return red, green, blue

    @staticmethod
    def _integer48(
            bytes48: bytes
    ) -> int:
        return bytes48[0] + (bytes48[1] << 8) + (bytes48[2] << 16) + (bytes48[3] << 24) + (bytes48[4] << 32) + (
                bytes48[5] << 40)


class TPCTextureFormat(IntEnum):
    Invalid = -1
    Greyscale = 0
    RGB = 1
    RGBA = 2
    DXT1 = 3
    DXT5 = 4
